# %%
import matplotlib.pyplot as plt
import numpy as np
from dtaidistance import dtw
from scipy.cluster.hierarchy import dendrogram, linkage

# testing if we can use DTW to align cell trajectories using a short reference pattern

np.random.seed(42)
timepoints = 50
cells = 8

# Create synthetic cell trajectories (e.g., PCA1 shape evolution)
cell_trajectories = [
    np.sin(np.linspace(0, 10, timepoints) + np.random.rand() * 2) for _ in range(cells)
]
# add extra transforms to signal
cell_trajectories[cells - 1] += np.sin(np.linspace(0, 5, timepoints)) * 2
cell_trajectories[cells - 2] += np.sin(np.linspace(0, 5, timepoints)) * 3

# %%
# plot cell trajectories
plt.figure(figsize=(8, 5))
for i in range(cells):
    plt.plot(cell_trajectories[i], label=f"Cell {i+1}")
plt.legend()
plt.title("Original Cell Trajectories")
plt.show()
# %%
# Set reference cell for all subsequent analysis
reference_cell = 0  # Use first cell as reference

# Compute DTW distance matrix
dtw_matrix = np.zeros((cells, cells))
for i in range(cells):
    for j in range(i + 1, cells):
        dtw_matrix[i, j] = dtw.distance(cell_trajectories[i], cell_trajectories[j])
        dtw_matrix[j, i] = dtw_matrix[i, j]

# Print distance matrix for examination
print("DTW Distance Matrix:")
for i in range(cells):
    print(f"Cell {i+1}: {dtw_matrix[reference_cell, i]:.2f}")

# Plot distance heatmap
plt.figure(figsize=(8, 6))
plt.imshow(dtw_matrix, cmap="viridis", origin="lower")
plt.colorbar(label="DTW Distance")
plt.title("DTW Distance Matrix Between Cells")
plt.xlabel("Cell Index")
plt.ylabel("Cell Index")
plt.tight_layout()
plt.show()

linkage_matrix = linkage(dtw_matrix, method="ward")

# Plot the dendrogram
plt.figure(figsize=(8, 5))
dendrogram(linkage_matrix, labels=[f"Cell {i+1}" for i in range(cells)])
plt.xlabel("Cells")
plt.ylabel("DTW Distance")
plt.title("Hierarchical Clustering of Cells Based on DTW")
plt.show()

# %%
# Align cells using DTW with distance filtering
# Set a threshold for maximum allowed DTW distance
# Cells with distances above this threshold won't be aligned
# This can be set based on the distribution of distances or domain knowledge
distance_threshold = np.median(dtw_matrix[reference_cell, :]) * 1.5  # Example threshold

print(f"Using distance threshold: {distance_threshold:.2f}")
print("Distances from reference cell:")
for i in range(cells):
    distance = dtw_matrix[reference_cell, i]
    status = (
        "Included"
        if distance <= distance_threshold or i == reference_cell
        else "Excluded (too dissimilar)"
    )
    print(f"Cell {i+1}: {distance:.2f} - {status}")

# Initialize aligned trajectories with the reference cell
aligned_cell_trajectories = [cell_trajectories[reference_cell].copy()]
alignment_status = [True]  # Reference cell is always included

for i in range(1, cells):
    distance = dtw_matrix[reference_cell, i]

    # Skip cells that are too dissimilar to the reference
    if distance > distance_threshold:
        aligned_cell_trajectories.append(
            np.full_like(cell_trajectories[reference_cell], np.nan)
        )
        alignment_status.append(False)
        continue

    # Find optimal warping path
    path = dtw.warping_path(cell_trajectories[reference_cell], cell_trajectories[i])

    # Create aligned trajectory by mapping query points to reference timeline
    aligned_trajectory = np.zeros_like(cell_trajectories[reference_cell])
    path_dict = {}

    # Group by reference indices
    for ref_idx, query_idx in path:
        if ref_idx not in path_dict:
            path_dict[ref_idx] = []
        path_dict[ref_idx].append(query_idx)

    # For each reference index, average the corresponding query values
    for ref_idx, query_indices in path_dict.items():
        query_values = [cell_trajectories[i][idx] for idx in query_indices]
        aligned_trajectory[ref_idx] = np.mean(query_values)

    aligned_cell_trajectories.append(aligned_trajectory)
    alignment_status.append(True)

# %%
# plot aligned cell trajectories (only included cells)
plt.figure(figsize=(10, 6))

# Plot reference cell first
plt.plot(aligned_cell_trajectories[0], "k-", linewidth=2.5, label=f"Reference (Cell 1)")

# Plot other cells that were successfully aligned
for i in range(1, cells):
    if alignment_status[i]:
        plt.plot(aligned_cell_trajectories[i], label=f"Cell {i+1}")

plt.legend()
plt.title("Aligned Cell Trajectories (Filtered by DTW Distance)")
plt.show()

# %%
# Visualize warping paths for examples
plt.figure(figsize=(15, 10))

# First find cells to include based on distance threshold
included_cells = [i for i in range(1, cells) if alignment_status[i]]
excluded_cells = [i for i in range(1, cells) if not alignment_status[i]]

# Show included cells examples
for idx, target_cell in enumerate(included_cells[: min(2, len(included_cells))]):
    plt.subplot(2, 3, idx + 1)

    # Get warping path
    path = dtw.warping_path(
        cell_trajectories[reference_cell], cell_trajectories[target_cell]
    )

    # Plot both signals
    plt.plot(cell_trajectories[reference_cell], label=f"Reference", linewidth=2)
    plt.plot(cell_trajectories[target_cell], label=f"Cell {target_cell+1}", linewidth=2)

    # Plot warping connections
    for ref_idx, query_idx in path:
        plt.plot(
            [ref_idx, query_idx],
            [
                cell_trajectories[reference_cell][ref_idx],
                cell_trajectories[target_cell][query_idx],
            ],
            "k-",
            alpha=0.1,
        )

    plt.title(
        f"Included - Cell {target_cell+1} (Dist: {dtw_matrix[reference_cell, target_cell]:.2f})"
    )
    plt.legend()

# Show excluded cells examples
for idx, target_cell in enumerate(excluded_cells[: min(2, len(excluded_cells))]):
    plt.subplot(2, 3, 3 + idx)

    plt.plot(cell_trajectories[reference_cell], label=f"Reference", linewidth=2)
    plt.plot(cell_trajectories[target_cell], label=f"Cell {target_cell+1}", linewidth=2)

    # Show distance value
    plt.title(
        f"Excluded - Cell {target_cell+1} (Dist: {dtw_matrix[reference_cell, target_cell]:.2f})"
    )
    plt.legend()

# Compare original and aligned for an included cell

if included_cells:
    plt.subplot(2, 3, 5)
    target_cell = included_cells[0]
    plt.plot(cell_trajectories[reference_cell], label="Reference", linewidth=2)
    plt.plot(
        cell_trajectories[target_cell],
        label=f"Original Cell {target_cell+1}",
        linewidth=2,
        linestyle="--",
        alpha=0.7,
    )
    plt.plot(
        aligned_cell_trajectories[target_cell],
        label=f"Aligned Cell {target_cell+1}",
        linewidth=2,
    )
    plt.title(f"Alignment Example (Included)")
    plt.legend()

# Show distance distribution
plt.subplot(2, 3, 6)
distances = dtw_matrix[reference_cell, 1:]  # Skip distance to self
plt.hist(distances, bins=10)s
plt.axvline(
    distance_threshold,
    color="r",
    linestyle="--",
    label=f"Threshold: {distance_threshold:.2f}",
)
plt.title("DTW Distance Distribution")
plt.xlabel("DTW Distance from Reference")
plt.ylabel("Count")
plt.legend()

plt.tight_layout()
plt.show()
# %%



# %%
import time

import numpy as np
import torch

from viscy.utils import cli_utils


def main():
    # create test data
    data = np.random.random((5, 5, 5, 128, 128))
    data_tensor = torch.Tensor(data)
    print(data_tensor.shape)

    # run feature logger on test data
    cli_utils.log_feature_map(
        data_tensor,
        "/home/christian.foley/virtual_staining/example_log/",
        dim_names=["batch", "channels"],
        spatial_dims=3,
        vmax=None,
    )


if __name__ == "main":
    start = time.time()
    try:
        main()
    except Exception as e:
        print(f"Error after {time.time() - start} seconds:\n{e}")



"""
Generate embeddings using the OpenPhenom model and save them to a zarr store
using VisCy Trainer and EmbeddingWriter callback.
"""

import importlib
import logging
import os
from pathlib import Path
from typing import Dict, List, Literal, Optional

import click
import torch
import yaml
from lightning.pytorch import LightningModule
from transformers import AutoModel

from viscy.data.triplet import TripletDataModule
from viscy.representation.embedding_writer import EmbeddingWriter
from viscy.trainer import VisCyTrainer


class OpenPhenomModule(LightningModule):
    def __init__(
        self,
        channel_reduction_methods: Optional[
            Dict[str, Literal["middle_slice", "mean", "max"]]
        ] = None,
        channel_names: Optional[List[str]] = None,
    ):
        """Initialize the OpenPhenom module.

        Parameters
        ----------
        channel_reduction_methods : dict, optional
            Dictionary mapping channel names to reduction methods:
            - "middle_slice": Take the middle slice along the depth dimension
            - "mean": Average across the depth dimension
            - "max": Take the maximum value across the depth dimension
        channel_names : list of str, optional
            List of channel names corresponding to the input channels

        Notes
        -----
        The module uses the OpenPhenom model from HuggingFace for generating embeddings.
        """
        super().__init__()

        self.channel_reduction_methods = channel_reduction_methods or {}
        self.channel_names = channel_names or []

        try:
            torch.set_float32_matmul_precision("high")
            self.model = AutoModel.from_pretrained(
                "recursionpharma/OpenPhenom", trust_remote_code=True
            )
            self.model.eval()
        except ImportError:
            raise ImportError(
                "Please install the OpenPhenom dependencies: "
                "pip install transformers"
            )

    def on_predict_start(self):
        # Move model to GPU when prediction starts
        self.model.to(self.device)

    def _reduce_5d_input(self, x: torch.Tensor) -> torch.Tensor:
        """Reduce 5D input (B, C, D, H, W) to 4D (B, C, H, W) using specified methods.

        Args:
            x: 5D input tensor

        Returns:
            4D tensor after applying reduction methods
        """
        if x.dim() != 5:
            return x

        B, C, D, H, W = x.shape
        result = torch.zeros((B, C, H, W), device=x.device)

        # Apply reduction method for each channel
        for c in range(C):
            channel_name = (
                self.channel_names[c] if c < len(self.channel_names) else f"channel_{c}"
            )
            # Default to middle slice if not specified
            method = self.channel_reduction_methods.get(channel_name, "middle_slice")

            if method == "middle_slice":
                result[:, c] = x[:, c, D // 2]
            elif method == "mean":
                result[:, c] = x[:, c].mean(dim=1)
            elif method == "max":
                result[:, c] = x[:, c].max(dim=1)[0]
            else:
                # Fallback to middle slice for unknown methods
                result[:, c] = x[:, c, D // 2]

        return result

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        """Extract features from the input images.

        Returns:
            Dictionary with features, projections (None), and index information
        """
        x = batch["anchor"]

        # OpenPhenom expects [B, C, H, W] but our data might be [B, C, D, H, W]
        # If 5D input, handle according to specified reduction methods
        if x.dim() == 5:
            x = self._reduce_5d_input(x)

        # Convert to uint8 as OpenPhenom expects uint8 inputs
        if x.dtype != torch.uint8:
            x = (
                ((x - x.min()) / (x.max() - x.min()) * 255)
                .clamp(0, 255)
                .to(torch.uint8)
            )

        # Get embeddings
        self.model.return_channelwise_embeddings = False
        features = self.model.predict(x)
        # Create empty projections tensor with same batch size as features
        # This ensures the EmbeddingWriter can process it
        projections = torch.zeros((features.shape[0], 0), device=features.device)

        return {
            "features": features,
            "projections": projections,
            "index": batch["index"],
        }


def load_config(config_file):
    """Load configuration from a YAML file."""
    with open(config_file, "r") as f:
        config = yaml.safe_load(f)
    return config


def load_normalization_from_config(norm_config):
    """Load a normalization transform from a configuration dictionary."""
    class_path = norm_config["class_path"]
    init_args = norm_config.get("init_args", {})

    # Split module and class name
    module_path, class_name = class_path.rsplit(".", 1)

    # Import the module
    module = importlib.import_module(module_path)

    # Get the class
    transform_class = getattr(module, class_name)

    # Instantiate the transform
    return transform_class(**init_args)


@click.command()
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    required=True,
    help="Path to YAML configuration file",
)
def main(config):
    """Extract OpenPhenom embeddings and save to zarr format using VisCy Trainer."""
    # Configure logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    # Load config file
    cfg = load_config(config)
    logger.info(f"Loaded configuration from {config}")

    # Prepare datamodule parameters
    dm_params = {}

    # Add data and tracks paths from the paths section
    if "paths" not in cfg:
        raise ValueError("Configuration must contain a 'paths' section")

    if "data_path" not in cfg["paths"]:
        raise ValueError(
            "Data path is required in the configuration file (paths.data_path)"
        )
    dm_params["data_path"] = cfg["paths"]["data_path"]

    if "tracks_path" not in cfg["paths"]:
        raise ValueError(
            "Tracks path is required in the configuration file (paths.tracks_path)"
        )
    dm_params["tracks_path"] = cfg["paths"]["tracks_path"]

    # Add datamodule parameters
    if "datamodule" not in cfg:
        raise ValueError("Configuration must contain a 'datamodule' section")

    # Prepare normalizations
    if (
        "normalizations" not in cfg["datamodule"]
        or not cfg["datamodule"]["normalizations"]
    ):
        raise ValueError(
            "Normalizations are required in the configuration file (datamodule.normalizations)"
        )

    norm_configs = cfg["datamodule"]["normalizations"]
    normalizations = [load_normalization_from_config(norm) for norm in norm_configs]
    dm_params["normalizations"] = normalizations

    # Copy all other datamodule parameters
    for param, value in cfg["datamodule"].items():
        if param != "normalizations":
            # Handle patch sizes
            if param == "patch_size":
                dm_params["initial_yx_patch_size"] = value
                dm_params["final_yx_patch_size"] = value
            else:
                dm_params[param] = value

    # Set up the data module
    logger.info("Setting up data module")
    dm = TripletDataModule(**dm_params)

    # Get model parameters for handling 5D inputs
    channel_reduction_methods = {}

    if "model" in cfg and "channel_reduction_methods" in cfg["model"]:
        channel_reduction_methods = cfg["model"]["channel_reduction_methods"]

    # Initialize OpenPhenom model with reduction settings
    logger.info("Loading OpenPhenom model")
    model = OpenPhenomModule(
        channel_reduction_methods=channel_reduction_methods,
        channel_names=dm_params.get("source_channel", []),
    )

    # Get dimensionality reduction parameters from config
    phate_kwargs = None
    pca_kwargs = None

    if "embedding" in cfg:
        if "phate_kwargs" in cfg["embedding"]:
            phate_kwargs = cfg["embedding"]["phate_kwargs"]
        if "pca_kwargs" in cfg["embedding"]:
            pca_kwargs = cfg["embedding"]["pca_kwargs"]
    # Check if output path exists and should be overwritten
    if "output_path" not in cfg["paths"]:
        raise ValueError(
            "Output path is required in the configuration file (paths.output_path)"
        )

    output_path = Path(cfg["paths"]["output_path"])
    output_dir = output_path.parent
    output_dir.mkdir(parents=True, exist_ok=True)

    overwrite = False
    if "execution" in cfg and "overwrite" in cfg["execution"]:
        overwrite = cfg["execution"]["overwrite"]
    elif output_path.exists():
        logger.warning(f"Output path {output_path} already exists, will overwrite")
        overwrite = True

    # Set up EmbeddingWriter callback
    embedding_writer = EmbeddingWriter(
        output_path=output_path,
        phate_kwargs=phate_kwargs,
        pca_kwargs=pca_kwargs,
        overwrite=overwrite,
    )

    # Set up and run VisCy trainer
    logger.info("Setting up VisCy trainer")
    trainer = VisCyTrainer(
        accelerator="gpu" if torch.cuda.is_available() else "cpu",
        devices=1,
        callbacks=[embedding_writer],
        inference_mode=True,
    )

    logger.info(f"Running prediction and saving to {output_path}")
    trainer.predict(model, datamodule=dm)

    # Save configuration if requested
    save_config_flag = True
    show_config_flag = True

    if "execution" in cfg:
        if "save_config" in cfg["execution"]:
            save_config_flag = cfg["execution"]["save_config"]
        if "show_config" in cfg["execution"]:
            show_config_flag = cfg["execution"]["show_config"]

    # Save configuration if requested
    if save_config_flag:
        config_path = os.path.join(output_dir, "config.yml")
        with open(config_path, "w") as f:
            yaml.dump(cfg, f, default_flow_style=False)
        logger.info(f"Configuration saved to {config_path}")

    # Display configuration if requested
    if show_config_flag:
        click.echo("\nConfiguration used:")
        click.echo("-" * 40)
        for key, value in cfg.items():
            click.echo(f"{key}:")
            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, list) and subkey == "normalizations":
                        click.echo(f"  {subkey}:")
                        for norm in subvalue:
                            click.echo(f"    - class_path: {norm['class_path']}")
                            click.echo(f"      init_args: {norm['init_args']}")
                    else:
                        click.echo(f"  {subkey}: {subvalue}")
            else:
                click.echo(f"  {value}")
        click.echo("-" * 40)

    logger.info("Done!")


if __name__ == "__main__":
    main()



# %% Imports and paths.
import timm
import torch
import torchview

from viscy.representation.contrastive import ContrastiveEncoder, StemDepthtoChannels
from viscy.representation.engine import ContrastiveModule

# %load_ext autoreload
# %autoreload 2
# %% Initialize the model and log the graph.
contra_model = ContrastiveEncoder(
    backbone="convnextv2_tiny",
    in_stack_depth=15,
    in_channels=2,
)  # other options: convnext_tiny resnet50
print(contra_model)

model_graph = torchview.draw_graph(
    contra_model,
    torch.randn(1, 2, 15, 224, 224),
    depth=3,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
model_graph.resize_graph(scale=2.5)
model_graph.visual_graph

# %% Initialize a resent50 model and log the graph.
contra_model = ContrastiveEncoder(
    backbone="resnet50", in_stack_depth=16, stem_kernel_size=(4, 3, 3)
)  # note that the resnet first layer takes 64 channels (so we can't have multiples of 3)
print(contra_model)
contra_model(torch.randn(1, 2, 16, 224, 224))
model_graph = torchview.draw_graph(
    contra_model,
    torch.randn(1, 2, 16, 224, 224),
    depth=3,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
model_graph.resize_graph(scale=2.5)
model_graph.visual_graph


# %% Initiatlize the lightning module and view the model.
contrastive_module = ContrastiveModule()
print(contrastive_module.encoder)

# %%
model_graph = torchview.draw_graph(
    contrastive_module.encoder,
    torch.randn(1, 2, 15, 200, 200),
    depth=3,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
model_graph.visual_graph

# %% Playground

available_models = timm.list_models(pretrained=True)

stem = StemDepthtoChannels(
    in_channels=2, out_channels=96, kernel_size=(5, 2, 2), in_stack_depth=15
)
print(stem)
stem_graph = torchview.draw_graph(
    stem,
    torch.randn(1, 2, 15, 256, 256),
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
stem_graph.visual_graph
# %%
encoder = timm.create_model(
    "convnext_tiny",
    pretrained=True,
    features_only=False,
    num_classes=200,
)

print(encoder)

# %%

encoder.stem = stem

model_graph = torchview.draw_graph(
    encoder,
    torch.randn(1, 2, 15, 256, 256),
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
model_graph.visual_graph
# %%
encoder.stem = torch.nn.Identity()

encoder_graph = torchview.draw_graph(
    encoder,
    torch.randn(1, 96, 128, 128),
    depth=2,  # adjust depth to zoom in.
    device="cpu",
)
# Print the image of the model.
encoder_graph.visual_graph

# %%



# %%
from pathlib import Path

import matplotlib.pyplot as plt
from monai.transforms import (
    CenterSpatialCropd,
    NormalizeIntensityd,
    RandAdjustContrastd,
    RandAffined,
    RandFlipd,
    RandGaussianNoised,
    RandGaussianSmoothd,
    RandScaleIntensityd,
)
from tqdm import tqdm

from viscy.data.ctmc_v1 import CTMCv1DataModule

# %%
channel = "DIC"
data_path = Path("/hpc/reference/imaging/ctmc")

normalize_transform = NormalizeIntensityd(keys=[channel], channel_wise=True)
crop_transform = CenterSpatialCropd(keys=[channel], roi_size=[1, 224, 224])

data = CTMCv1DataModule(
    train_data_path=data_path / "CTMCV1_test.zarr",
    val_data_path=data_path / "CTMCV1_train.zarr",
    train_transforms=[
        normalize_transform,
        RandAffined(
            keys=[channel],
            rotate_range=[3.14, 0.0, 0.0],
            scale_range=[0.0, [-0.6, 0.1], [-0.6, 0.1]],
            prob=0.8,
            padding_mode="zeros",
        ),
        RandFlipd(keys=[channel], prob=0.5, spatial_axis=(1, 2)),
        RandAdjustContrastd(keys=[channel], prob=0.5, gamma=(0.8, 1.2)),
        RandScaleIntensityd(keys=[channel], factors=0.3, prob=0.5),
        RandGaussianNoised(keys=[channel], prob=0.5, mean=0.0, std=0.2),
        RandGaussianSmoothd(
            keys=[channel],
            sigma_x=(0.05, 0.3),
            sigma_y=(0.05, 0.3),
            sigma_z=(0.05, 0.0),
            prob=0.5,
        ),
        crop_transform,
    ],
    val_transforms=[normalize_transform, crop_transform],
    batch_size=32,
    num_workers=0,
    channel_name=channel,
)

# %%
data.setup("fit")
dmt = data.train_dataloader()
dmv = data.val_dataloader()

# %%
for batch in tqdm(dmt):
    img = batch["source"]
    img[:, :, :, 32:64, 32:64] = 0
    f, ax = plt.subplots(5, 5, figsize=(15, 15))
    for sample, a in zip(img, ax.flatten()):
        a.imshow(sample[0, 0].cpu().numpy(), cmap="gray", vmin=-5, vmax=5)
        a.axis("off")
    f.tight_layout()
    break

# %%
for batch in tqdm(dmv):
    img = batch["source"]
    f, ax = plt.subplots(5, 5, figsize=(15, 15))
    for sample, a in zip(img, ax.flatten()):
        a.imshow(sample[0, 0].cpu().numpy(), cmap="gray", vmin=-5, vmax=5)
        a.axis("off")
    f.tight_layout()
    break


# %%



import torch
import torch.nn as nn

from viscy.unet.networks.layers.ConvBlock2D import ConvBlock2D


class Unet2d(nn.Module):
    def __name__(self):
        return "Unet2d"

    def __init__(
        self,
        in_channels=1,
        out_channels=1,
        kernel_size=(3, 3),
        residual=False,
        dropout=0.2,
        num_blocks=4,
        num_block_layers=2,
        num_filters=[],
        task="seg",
    ):
        """
        2D Unet with variable input/output channels and depth (block numbers).
        Follows 2D UNet Architecture:
            1) Unet: https://arxiv.org/pdf/1505.04597.pdf
            2) residual Unet: https://arxiv.org/pdf/1711.10684.pdf

        :param int in_channels: number of feature channels in
        :param int out_channels: number of feature channels out
        :param int/tuple(int,int) kernel_size: size of x and y dimensions
            of conv kernels in blocks
        :param bool residual: see name
        :param float dropout: probability of dropout, between 0 and 0.5
        :param int num_blocks: number of convolutional blocks on encoder and decoder
        :param int num_block_layers: number of layers per block
        :param list[int] num_filters: list of filters/feature levels
            at each conv block depth
        :param str task: network task (for virtual staining this is regression),
            one of 'seg','reg'
        """

        super(Unet2d, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.residual = residual
        self.dropout = dropout
        self.num_blocks = num_blocks
        self.num_block_layers = num_block_layers
        self.task = task

        # ----- set static parameters -----#
        self.block_padding = "same"
        down_mode = "avgpool"  # TODO set static avgpool
        up_mode = "bilinear"  # TODO set static bilinear
        activation = "relu"  # TODO set static relu
        self.bottom_block_spatial = False  # TODO set static

        # ----- Standardize Filter Sequence -----#
        if len(num_filters) != 0:
            assert len(num_filters) == num_blocks + 1, (
                "Length of num_filters must be equal to num_blocks + 1 "
                "(number of convolutional blocks per path)."
            )
            self.num_filters = num_filters
        else:
            self.num_filters = [pow(2, i) * 16 for i in range(num_blocks + 1)]
        downsampling_filters = [in_channels] + self.num_filters
        upsampling_filters = [
            self.num_filters[-(i + 1)] + self.num_filters[-(i + 2)]
            for i in range(len(self.num_filters))
            if i < len(self.num_filters) - 1
        ] + [out_channels]

        # ----- Downsampling steps -----#
        self.down_list = []
        if down_mode == "maxpool":
            for i in range(num_blocks):
                self.down_list.append(nn.MaxPool2d(kernel_size=2))
        elif down_mode == "avgpool":
            for i in range(num_blocks):
                self.down_list.append(nn.AvgPool2d(kernel_size=2))
        elif down_mode == "conv":
            raise NotImplementedError("Not yet implemented!")
            # TODO: implement.
        self.register_modules(self.down_list, "down_samp")

        # ----- Upsampling steps -----#
        self.up_list = []
        if up_mode == "bilinear":
            for i in range(num_blocks):
                self.up_list.append(
                    nn.Upsample(mode=up_mode, scale_factor=2, align_corners=False)
                )
        elif up_mode == "conv":
            raise NotImplementedError("Not yet implemented!")
            # TODO: implement
        elif up_mode == "tconv":
            raise NotImplementedError("Not yet implemented!")
            # TODO: implement
        else:
            raise NotImplementedError(f"Upsampling mode '{up_mode}' not supported.")

        # ----- Convolutional blocks -----#
        self.down_conv_blocks = []
        for i in range(num_blocks):
            self.down_conv_blocks.append(
                ConvBlock2D(
                    downsampling_filters[i],
                    downsampling_filters[i + 1],
                    dropout=self.dropout,
                    residual=self.residual,
                    activation=activation,
                    kernel_size=self.kernel_size,
                    num_repeats=self.num_block_layers,
                )
            )
        self.register_modules(self.down_conv_blocks, "down_conv_block")

        self.bottom_transition_block = ConvBlock2D(
            self.num_filters[-2],
            self.num_filters[-1],
            dropout=self.dropout,
            residual=self.residual,
            activation=activation,
            kernel_size=self.kernel_size,
            num_repeats=self.num_block_layers,
        )

        self.up_conv_blocks = []
        for i in range(num_blocks):
            self.up_conv_blocks.append(
                ConvBlock2D(
                    upsampling_filters[i],
                    downsampling_filters[-(i + 2)],
                    dropout=self.dropout,
                    residual=self.residual,
                    activation=activation,
                    kernel_size=self.kernel_size,
                    num_repeats=self.num_block_layers,
                )
            )
        self.register_modules(self.up_conv_blocks, "up_conv_block")

        # ----- Terminal Block and Activation Layer -----#
        if self.task == "reg":
            self.terminal_block = ConvBlock2D(
                downsampling_filters[1],
                out_channels,
                dropout=self.dropout,
                residual=False,
                activation="linear",
                num_repeats=1,
                norm="none",
                kernel_size=self.kernel_size,
            )
        else:
            self.terminal_block = ConvBlock2D(
                downsampling_filters[1],
                out_channels,
                dropout=self.dropout,
                residual=False,
                activation=activation,
                num_repeats=1,
                norm="none",
                kernel_size=self.kernel_size,
            )

    def forward(self, x, validate_input=False):
        """
        Forward call of network
            - x -> Torch.tensor: input image stack

        Call order:
            => num_block 2D convolutional blocks, with downsampling in between (encoder)
            => num_block 2D convolutional blocks, with upsampling between them (decoder)
            => skip connections between corresponding blocks on encoder and decoder
            => terminal block collapses to output dimensions

        :param torch.tensor x: input image
        :param bool validate_input: Deactivates assertions which are redundant
            if forward pass is being traced by tensorboard writer.
        """
        # handle input exceptions
        if validate_input:
            assert x.shape[-1] == x.shape[-2], "Input must be square in xy"
            assert x.shape[-3] == self.in_channels, (
                f"Input channels must equal network input channels: {self.in_channels}"
            )
        x = x.squeeze(2)
        # encoder
        skip_tensors = []
        for i in range(self.num_blocks):
            x = self.down_conv_blocks[i](x, validate_input=validate_input)
            skip_tensors.append(x)
            x = self.down_list[i](x)

        # transition block
        x = self.bottom_transition_block(x)

        # decoder
        for i in range(self.num_blocks):
            x = self.up_list[i](x)
            x = torch.cat([x, skip_tensors[-1 * (i + 1)]], 1)
            x = self.up_conv_blocks[i](x, validate_input=validate_input)

        # output channel collapsing layer
        x = self.terminal_block(x)

        return x.unsqueeze(2)

    def register_modules(self, module_list, name):
        """
        Helper function that registers modules stored in a list to the model object
        so that they can be seen by PyTorch optimizer.

        Used to enable model graph creation with
        non-sequential model types and dynamic layer numbers

        :param list(torch.nn.module) module_list: list of modules to register
        :param str name: name of module type
        """
        for i, module in enumerate(module_list):
            self.add_module(f"{name}_{str(i)}", module)



import os
import unittest
import warnings

import nose.tools
import numpy as np
import numpy.testing
import pandas as pd
import skimage.io as sk_im_io
from testfixtures import TempDirectory

from viscy.preprocessing.generate_masks import MaskProcessor
from viscy.utils import aux_utils as aux_utils


class TestMaskProcessor(unittest.TestCase):
    def setUp(self):
        """Set up a directory for mask generation,"""

        self.tempdir = TempDirectory()
        self.temp_path = self.tempdir.path
        self.meta_fname = "frames_meta.csv"
        frames_meta = aux_utils.make_dataframe()

        # create an image with bimodal hist
        x = np.linspace(-4, 4, 32)
        y = x.copy()
        z = np.linspace(-3, 3, 8)
        xx, yy, zz = np.meshgrid(x, y, z)
        sph = xx**2 + yy**2 + zz**2
        fg = (sph <= 8) * (8 - sph)
        fg[fg > 1e-8] = (fg[fg > 1e-8] / np.max(fg)) * 127 + 128
        fg = np.around(fg).astype("uint8")
        bg = np.around((sph > 8) * sph).astype("uint8")
        object1 = fg + bg

        # create an image with a rect
        rec = np.zeros(sph.shape)
        rec[3:30, 14:18, 3:6] = 120
        rec[14:18, 3:30, 3:6] = 120

        self.sph_object = object1
        self.rec_object = rec

        self.channel_ids = [1, 2]
        self.time_ids = 0
        self.pos_ids = 1
        self.int2str_len = 3

        for z in range(sph.shape[2]):
            im_name = aux_utils.get_im_name(
                time_idx=self.time_ids,
                channel_idx=1,
                slice_idx=z,
                pos_idx=self.pos_ids,
            )
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                sk_im_io.imsave(
                    os.path.join(self.temp_path, im_name),
                    object1[:, :, z].astype("uint8"),
                )
            frames_meta = frames_meta.append(
                aux_utils.parse_idx_from_name(im_name=im_name, dir_name=self.temp_path),
                ignore_index=True,
            )
        for z in range(rec.shape[2]):
            im_name = aux_utils.get_im_name(
                time_idx=self.time_ids,
                channel_idx=2,
                slice_idx=z,
                pos_idx=self.pos_ids,
            )
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                sk_im_io.imsave(
                    os.path.join(self.temp_path, im_name),
                    rec[:, :, z].astype("uint8"),
                )
            frames_meta = frames_meta.append(
                aux_utils.parse_idx_from_name(im_name=im_name, dir_name=self.temp_path),
                ignore_index=True,
            )
        # Write metadata
        frames_meta.to_csv(os.path.join(self.temp_path, self.meta_fname), sep=",")

        self.output_dir = os.path.join(self.temp_path, "mask_dir")
        self.mask_gen_inst = MaskProcessor(
            input_dir=self.temp_path,
            output_dir=self.output_dir,
            channel_ids=self.channel_ids,
        )

    def tearDown(self):
        """Tear down temporary folder and file structure"""
        TempDirectory.cleanup_all()
        self.assertFalse(os.path.isdir(self.temp_path))

    def test_init(self):
        """Test init"""
        self.assertEqual(self.mask_gen_inst.input_dir, self.temp_path)
        self.assertEqual(self.mask_gen_inst.output_dir, self.output_dir)
        nose.tools.assert_equal(self.mask_gen_inst.mask_channel, 3)
        nose.tools.assert_equal(
            self.mask_gen_inst.mask_dir,
            os.path.join(self.output_dir, "mask_channels_1-2"),
        )
        self.assertListEqual(self.channel_ids, self.channel_ids)
        nose.tools.assert_equal(self.mask_gen_inst.nested_id_dict, None)

    def test_get_mask_dir(self):
        """Test get_mask_dir"""
        mask_dir = os.path.join(self.output_dir, "mask_channels_1-2")
        nose.tools.assert_equal(self.mask_gen_inst.get_mask_dir(), mask_dir)

    def test_get_mask_channel(self):
        """Test get_mask_channel"""
        nose.tools.assert_equal(self.mask_gen_inst.get_mask_channel(), 3)

    def test_generate_masks_uni(self):
        """Test generate masks"""
        self.mask_gen_inst.generate_masks(str_elem_radius=1)
        frames_meta = pd.read_csv(
            os.path.join(self.mask_gen_inst.get_mask_dir(), "frames_meta.csv"),
            index_col=0,
        )
        # 8 slices and 3 channels
        exp_len = 8
        nose.tools.assert_equal(len(frames_meta), exp_len)
        for idx in range(exp_len):
            nose.tools.assert_equal(
                "im_c003_z00{}_t000_p001.npy".format(idx),
                frames_meta.iloc[idx]["file_name"],
            )

    def test_generate_masks_nonuni(self):
        """Test generate_masks with non-uniform structure"""
        rec = self.rec_object[:, :, 3:6]
        channel_ids = 0
        time_ids = 0
        pos_ids = [1, 2]
        frames_meta = aux_utils.make_dataframe()

        for z in range(self.sph_object.shape[2]):
            im_name = aux_utils.get_im_name(
                time_idx=time_ids,
                channel_idx=channel_ids,
                slice_idx=z,
                pos_idx=pos_ids[0],
            )
            sk_im_io.imsave(
                os.path.join(self.temp_path, im_name),
                self.sph_object[:, :, z].astype("uint8"),
            )
            frames_meta = frames_meta.append(
                aux_utils.parse_idx_from_name(im_name=im_name, dir_name=self.temp_path),
                ignore_index=True,
            )
        for z in range(rec.shape[2]):
            im_name = aux_utils.get_im_name(
                time_idx=time_ids,
                channel_idx=channel_ids,
                slice_idx=z,
                pos_idx=pos_ids[1],
            )
            sk_im_io.imsave(
                os.path.join(self.temp_path, im_name), rec[:, :, z].astype("uint8")
            )
            frames_meta = frames_meta.append(
                aux_utils.parse_idx_from_name(im_name=im_name, dir_name=self.temp_path),
                ignore_index=True,
            )
        # Write metadata
        frames_meta.to_csv(os.path.join(self.temp_path, self.meta_fname), sep=",")

        self.output_dir = os.path.join(self.temp_path, "mask_dir")
        mask_gen_inst = MaskProcessor(
            input_dir=self.temp_path,
            output_dir=self.output_dir,
            channel_ids=channel_ids,
            uniform_struct=False,
        )
        exp_nested_id_dict = {0: {0: {1: [0, 1, 2, 3, 4, 5, 6, 7], 2: [0, 1, 2]}}}
        numpy.testing.assert_array_equal(
            mask_gen_inst.nested_id_dict[0][0][1], exp_nested_id_dict[0][0][1]
        )
        numpy.testing.assert_array_equal(
            mask_gen_inst.nested_id_dict[0][0][2], exp_nested_id_dict[0][0][2]
        )

        mask_gen_inst.generate_masks(str_elem_radius=1)

        frames_meta = pd.read_csv(
            os.path.join(mask_gen_inst.get_mask_dir(), "frames_meta.csv"),
            index_col=0,
        )
        # pos1: 8 slices, pos2: 3 slices
        exp_len = 8 + 3
        nose.tools.assert_equal(len(frames_meta), exp_len)
        mask_fnames = frames_meta["file_name"].tolist()
        exp_mask_fnames = [
            "im_c001_z000_t000_p001.npy",
            "im_c001_z000_t000_p002.npy",
            "im_c001_z001_t000_p001.npy",
            "im_c001_z001_t000_p002.npy",
            "im_c001_z002_t000_p001.npy",
            "im_c001_z002_t000_p002.npy",
            "im_c001_z003_t000_p001.npy",
            "im_c001_z004_t000_p001.npy",
            "im_c001_z005_t000_p001.npy",
            "im_c001_z006_t000_p001.npy",
            "im_c001_z007_t000_p001.npy",
        ]
        nose.tools.assert_list_equal(mask_fnames, exp_mask_fnames)



# %% script to compare the output from the supervised model and human revised annotations to get the accuracy of the model

import numpy as np
from iohub import open_ome_zarr
from scipy.ndimage import label

# %% datapaths

# Path to model output
data_out_path = "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/supervised_inf_pred/supervised_test.zarr"

# Path to the human revised annotations
human_corrected_path = "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/supervised_inf_pred/supervised_test_corrected.zarr"

# %% Load data and compute the number of objects in each class

data_out = open_ome_zarr(data_out_path, layout="hcs", mode="r+")
human_corrected = open_ome_zarr(human_corrected_path, layout="hcs", mode="r+")

out_medians = []
HC_medians = []
for well_id, well_data in data_out.wells():
    well_name, well_no = well_id.split("/")

    for pos_name, pos_data in well_data.positions():

        out_data = pos_data.data.numpy()
        T, C, Z, Y, X = out_data.shape

        HC_data = human_corrected[well_id + "/" + pos_name + "/0"]
        HC_data = HC_data.numpy()

        # Compute the number of objects in the model output
        for t in range(T):
            out_img = out_data[t, 0, 0]

            # Compute the number of objects in the model output
            out_labeled, num_out_objects = label(out_img > 0)

            # Compute the median of pixel values in each object in the model output
            for obj_id in range(1, num_out_objects + 1):
                obj_pixels = out_img[out_labeled == obj_id]
                out_medians.append(np.median(obj_pixels))

            # repeat for human acorrected annotations
            HC_img = HC_data[t, 0, 0]
            HC_labeled, num_HC_objects = label(HC_img > 0)

            for obj_id in range(1, num_HC_objects + 1):
                obj_pixels = HC_img[HC_labeled == obj_id]
                HC_medians.append(np.median(obj_pixels))

# %% Compute the accuracy

num_twos_in_out_medians = out_medians.count(2)
num_twos_in_HC_medians = HC_medians.count(2)
error_inf = (
    (num_twos_in_HC_medians - num_twos_in_out_medians) / num_twos_in_HC_medians
) * 100

num_ones_in_out_medians = out_medians.count(1)
num_ones_in_HC_medians = HC_medians.count(1)
error_uninf = (
    (num_ones_in_HC_medians - num_ones_in_out_medians) / num_ones_in_HC_medians
) * 100

avg_error = (np.abs(error_inf) + np.abs(error_uninf)) / 2

accuracy = 100 - avg_error

# %%



import json
from pathlib import Path

from viscy.utils.aux_utils import read_config


def test_read_config(tmp_path: Path):
    config = tmp_path / "config.yml"
    # The function doesn't care about file format, names just have to start with im_
    test_config = {"param": 10}
    config.write_text(json.dumps(test_config))
    config = read_config(str(config))
    assert config == test_config



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


class ConvBlock2D(nn.Module):
    def __init__(
        self,
        in_filters,
        out_filters,
        dropout=False,
        norm="batch",
        residual=True,
        activation="relu",
        transpose=False,
        kernel_size=3,
        num_repeats=3,
        filter_steps="first",
        layer_order="can",
    ):
        """
        Convolutional block for lateral layers in Unet

        Format for layer initialization is as follows:
            if layer type specified
            => for number of layers
            => add layer to list of that layer type
            => register elements of list
        This is done to allow for dynamic layer number specification in the conv blocks,
        which allows us to change the parameter numbers of the network.

        :param int in_filters: number of images in in stack
        :param int out_filters: number of images in out stack
        :param float dropout: dropout probability (False => 0)
        :param str norm: normalization type: 'batch', 'instance'
        :param bool residual: as name
        :param str activation: activation function: 'relu', 'leakyrelu', 'elu', 'selu'
        :param bool transpose: as name
        :param int/tuple kernel_size: convolutional kernel size
        :param int num_repeats: number of times the layer_order layer sequence
            is repeated in the block
        :param str filter_steps: determines where in the block
            the filters inflate channels (learn abstraction information):
            'linear','first','last'
        :param str layer_order: order of conv, norm, and act layers in block:
            'can', 'cna', 'nca', etc
        """

        super(ConvBlock2D, self).__init__()
        self.in_filters = in_filters
        self.out_filters = out_filters
        self.dropout = dropout
        self.norm = norm
        self.residual = residual
        self.activation = activation
        self.transpose = transpose
        self.num_repeats = num_repeats
        self.filter_steps = filter_steps
        self.layer_order = layer_order

        # ---- Handle Kernel ----#
        ks = kernel_size
        if isinstance(ks, int):
            assert ks % 2 == 1, "Kernel dims must be odd"
        elif isinstance(ks, tuple):
            for i in range(len(ks)):
                assert ks[i] % 2 == 1, "Kernel dims must be odd"
            assert i == 1, "kernel_size length must be 2"
        else:
            raise AttributeError("'kernel_size' must be either int or tuple")
        self.kernel_size = kernel_size

        # ----- Init Dropout -----#
        if self.dropout:
            self.drop_list = []
            for i in range(self.num_repeats):
                self.drop_list.append(nn.Dropout2d(int(self.dropout)))

        # ---- Init linear filter steps ----#
        steps = np.linspace(in_filters, out_filters, num_repeats + 1).astype(int)

        # ----- Init Normalization Layers -----#
        # The parameters governing the initiation logic flow are:
        #                 self.norm
        #                 self.num_repeats
        #                 self.filter_steps
        self.norm_list = [None for i in range(num_repeats)]
        if self.norm == "batch":
            for i in range(self.num_repeats):
                if self.filter_steps == "linear":
                    self.norm_list[i] = nn.BatchNorm2d(steps[i + 1])
                elif self.filter_steps == "first":
                    self.norm_list[i] = nn.BatchNorm2d(steps[-1])
                elif self.filter_steps == "last":
                    if i < self.num_repeats - 1:
                        self.norm_list[i] = nn.BatchNorm2d(steps[0])
                    else:
                        self.norm_list[i] = nn.BatchNorm2d(steps[-1])
        elif self.norm == "instance":
            for i in range(self.num_repeats):
                if self.filter_steps == "linear":
                    self.norm_list[i] = nn.InstanceNorm2d(steps[i + 1])
                elif self.filter_steps == "first":
                    self.norm_list[i] = nn.InstanceNorm2d(steps[-1])
                elif self.filter_steps == "last":
                    if i < self.num_repeats - 1:
                        self.norm_list[i] = nn.InstanceNorm2d(steps[0])
                    else:
                        self.norm_list[i] = nn.InstanceNorm2d(steps[-1])
        self.register_modules(self.norm_list, f"{norm}_norm")

        # ----- Init Conv Layers -----#
        # init conv layers and determine transposition during convolution
        # The parameters governing the initiation logic flow are:
        #                 self.transpose
        #                 self.num_repeats
        #                 self.filter_steps
        # See above for definitions.
        # -------#

        self.conv_list = []
        if self.filter_steps == "linear":  # learn progressively over steps
            for i in range(self.num_repeats):
                depth_pair = (
                    (steps[i], steps[i + 1])
                    if i + 1 < num_repeats
                    else (steps[i], steps[-1])
                )
                if self.transpose:
                    self.conv_list.append(
                        nn.ConvTranspose2d(
                            depth_pair[0],
                            depth_pair[1],
                            kernel_size=kernel_size,
                            padding="same",
                        )
                    )
                else:
                    self.conv_list.append(
                        nn.Conv2d(
                            depth_pair[0],
                            depth_pair[1],
                            kernel_size=kernel_size,
                            padding="same",
                        )
                    )

        elif self.filter_steps == "first":  # learn in the first convolution
            if self.transpose:
                raise NotImplementedError(
                    "PyTorch-side problem with 'same' padding in ConvTranspose2d."
                )
                for i in range(self.num_repeats):
                    if i == 0:
                        self.conv_list.append(
                            nn.ConvTranspose2d(
                                in_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
                    else:
                        self.conv_list.append(
                            nn.ConvTranspose2d(
                                out_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
            else:
                for i in range(self.num_repeats):
                    if i == 0:
                        self.conv_list.append(
                            nn.Conv2d(
                                in_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
                    else:
                        self.conv_list.append(
                            nn.Conv2d(
                                out_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )

        elif self.filter_steps == "last":  # learn in the last convolution
            if self.transpose:
                raise NotImplementedError(
                    "Problem with 'same' padding in ConvTranspose2d."
                )
                for i in range(self.num_repeats):
                    if i == self.num_repeats - 1:
                        self.conv_list.append(
                            nn.ConvTranspose2d(
                                in_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
                    else:
                        self.conv_list.append(
                            nn.ConvTranspose2d(
                                out_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
            else:
                for i in range(self.num_repeats):
                    if i == self.num_repeats - 1:
                        self.conv_list.append(
                            nn.Conv2d(
                                in_filters,
                                out_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
                    else:
                        self.conv_list.append(
                            nn.Conv2d(
                                in_filters,
                                in_filters,
                                kernel_size=kernel_size,
                                padding="same",
                            )
                        )
        self.register_modules(self.conv_list, "Conv2d")

        # ----- Init Residual Layer -----#
        self.resid_conv = nn.Conv2d(
            self.in_filters, self.out_filters, kernel_size=1, padding=0
        )

        # ----- Init Activation Layers -----#
        self.act_list = []
        if self.activation == "relu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.ReLU())
        elif self.activation == "leakyrelu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.LeakyReLU())
        elif self.activation == "elu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.ELU())
        elif self.activation == "selu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.SELU())
        elif self.activation != "linear":
            raise NotImplementedError(
                f"Activation type {self.activation} not supported."
            )
        self.register_modules(self.act_list, f"{self.activation}_act")

    def forward(self, x, validate_input=False):
        """
        Forward call of convolutional block

        Order of layers within the block is defined by the 'layer_order' parameter,
        which is a string of 'c's, 'a's and 'n's
        in reference to convolution, activation, and normalization layers.
        This sequence is repeated num_repeats times.

        Recommended layer order:   convolution -> activation -> normalization

        Regardless of layer order,
        the final layer sequence in the block always ends in activation.
        This allows for usage of passthrough layers
        or a final output activation function determined separately.

        Residual blocks:
            if input channels are greater than output channels,
            we use a 1x1 convolution on input to get desired feature channels;
            if input channels are less than output channels,
            we zero-pad input channels to output channel size.

        :param torch.tensor x: input tensor
        :param bool validate_input: Deactivates assertions
            which are redundant if forward pass is being traced by tensorboard writer.
        """
        if validate_input:
            if isinstance(self.kernel_size, int):
                assert (
                    x.shape[-1] > self.kernel_size and x.shape[-2] > self.kernel_size
                ), (
                    f"Input size"
                    f" {x.shape} too small for kernel of size {self.kernel_size}"
                )
            elif isinstance(self.kernel_size, tuple):
                assert (
                    x.shape[-1] > self.kernel_size[-1]
                    and x.shape[-2] > self.kernel_size[-2]
                ), (
                    f"Input size"
                    f" {x.shape} too small for kernel of size {self.kernel_size}"
                )

        x_0 = x
        for i in range(self.num_repeats):
            order = list(self.layer_order)
            while len(order) > 0:
                layer = order.pop(0)
                if layer == "c":
                    x = self.conv_list[i](x)
                    if self.dropout:
                        x = self.drop_list[i](x)
                elif layer == "a":
                    if i < self.num_repeats - 1 or self.activation != "linear":
                        x = self.act_list[i](x)
                elif layer == "n" and self.norm_list[i]:
                    x = self.norm_list[i](x)

        # residual summation after final activation/normalization
        if self.residual:
            if self.in_filters > self.out_filters:
                x_0 = self.resid_conv(x_0)
            elif self.in_filters < self.out_filters:
                x_0 = F.pad(
                    x_0,
                    (*[0] * 4, self.out_filters - self.in_filters, *[0] * 3),
                    mode="constant",
                    value=0,
                )
            x = torch.add(x_0, x)

        return x

    def model(self):
        """
        Allows calling of parameters inside ConvBlock object:
        'ConvBlock.model().parameters()''

        Layer order:       convolution -> normalization -> activation

        We can make a list of layer modules and unpack them into nn.Sequential.
        Note: this is distinct from the forward call
        because we want to use the forward call with addition,
        since this is a residual block.
        The forward call performs the residial calculation,
        and all the parameters can be seen by the optimizer when given this model.
        """
        layers = []

        for i in range(self.num_repeats):
            layers.append(self.conv_list[i])
            if self.dropout:
                layers.append(self.drop_list[i])
            if self.norm[i]:
                layers.append(self.norm_list[i])
            if i < len(self.act_list):
                layers.append(self.act_list[i])

        return nn.Sequential(*layers)

    def register_modules(self, module_list, name):
        """
        Helper function that registers modules stored in a list to the model object
        so that they can be seen by PyTorch optimizer.

        Used to enable model graph creation
        with non-sequential model types and dynamic layer numbers

        :param list(torch.nn.module) module_list: list of modules to register
        :param str name: name of module type
        """
        for i, module in enumerate(module_list):
            self.add_module(f"{name}_{str(i)}", module)



from numpy.testing import assert_allclose

from viscy.preprocessing.pixel_ratio import sematic_class_weights


def test_sematic_class_weights(small_hcs_dataset):
    weights = sematic_class_weights(small_hcs_dataset, "GFP")
    assert weights.shape == (3,)
    assert_allclose(weights[0], 1.0, atol=1e-5)
    # infinity
    assert weights[1] > 1.0
    assert weights[2] > 1.0
    assert sematic_class_weights(small_hcs_dataset, "GFP", num_classes=2).shape == (2,)



# metrics for the knowledge distillation figure

# %%
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from sklearn.metrics import accuracy_score, classification_report, f1_score

# %%
# Mantis
test_virus = ["C/2/000000", "C/2/001001"]
test_mock = ["B/3/000000", "B/3/000001"]

# Mantis
TRAIN_FOVS = ["C/2/000001", "C/2/001000", "B/3/001000", "B/3/001001"]

VAL_FOVS = test_virus + test_mock

# %%
prediction_from_scratch = pd.read_csv(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/bootstrap-labels/test/from-scratch-last-1126.csv"
)
prediction_from_scratch["pretraining"] = "ImageNet"

prediction_finetuned = pd.read_csv(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/bootstrap-labels/test/fine-tune-last-1126.csv"
)
pretrained_name = "DynaCLR"
prediction_finetuned["pretraining"] = pretrained_name

prediction = pd.concat([prediction_from_scratch, prediction_finetuned], axis=0)

# %%
prediction = prediction[prediction["fov_name"].isin(VAL_FOVS)]
prediction["prediction_binary"] = prediction["prediction"] > 0.5
prediction

# %%
print(
    classification_report(
        prediction["label"], prediction["prediction_binary"], digits=3
    )
)

# %%
prediction["HPI"] = prediction["t"] / 6 + 3

bins = [3, 6, 9, 12, 15, 18, 21, 24]
labels = [f"{start}-{end}" for start, end in zip(bins[:-1], bins[1:])]
prediction["stage"] = pd.cut(prediction["HPI"], bins=bins, labels=labels, right=True)
prediction["well"] = prediction["fov_name"].apply(
    lambda x: "ZIKV" if x in test_virus else "Mock"
)
comparison = prediction.melt(
    id_vars=["fov_name", "id", "HPI", "well", "stage", "pretraining"],
    value_vars=["label", "prediction_binary"],
    var_name="source",
    value_name="value",
)
with sns.axes_style("whitegrid"):
    ax = sns.lineplot(
        data=comparison[comparison["pretraining"] == pretrained_name],
        x="HPI",
        y="value",
        hue="well",
        hue_order=["Mock", "ZIKV"],
        style="source",
        errorbar=None,
        color="gray",
    )
    ax.set_ylabel("Infection ratio")

# %%
id_vars = ["stage", "pretraining"]

accuracy_by_t = prediction.groupby(id_vars).apply(
    lambda x: float(accuracy_score(x["label"], x["prediction_binary"]))
)
f1_by_t = prediction.groupby(id_vars).apply(
    lambda x: float(f1_score(x["label"], x["prediction_binary"]))
)

metrics_df = pd.DataFrame(
    data={"accuracy": accuracy_by_t.values, "F1": f1_by_t.values},
    index=f1_by_t.index,
).reset_index()

metrics_long = metrics_df.melt(
    id_vars=id_vars,
    value_vars=["accuracy"],
    var_name="metric",
    value_name="score",
)

with sns.axes_style("ticks"):
    plt.style.use("../figures/figure.mplstyle")
    g = sns.catplot(
        data=metrics_long,
        x="stage",
        y="score",
        hue="pretraining",
        kind="point",
        linewidth=1.5,
        linestyles="--",
    )
    g.set_axis_labels("HPI", "accuracy")
    sns.move_legend(g, "upper left", bbox_to_anchor=(0.35, 1.1))
    g.figure.set_size_inches(3.5, 1.5)
    g.set(xlim=(-1, 7), ylim=(0.6, 1.0))
    plt.show()


# %%
g.figure.savefig(
    Path.home()
    / "gdrive/publications/dynaCLR/2025_dynaCLR_paper/fig_manuscript_svg/figure_knowledge_distillation/figure_parts/accuracy_students.pdf",
    dpi=300,
)

# %%



# %% [markdown]
"""
# 3D Virtual Staining of Neuromast
---
## Prediction using the VSNeuromast to predict nuclei and membrane from phase.
This example shows how to virtually stain zebrafish neuromast cells using the _VSNeuromast_ model.
The model is trained to predict the membrane and nuclei channels from the phase channel.
"""
# %% Imports and paths
from pathlib import Path

from iohub import open_ome_zarr
from plot import plot_vs_n_fluor
from viscy.data.hcs import HCSDataModule

# Viscy classes for the trainer and model
from viscy.translation.engine import VSUNet
from viscy.translation.predict_writer import HCSPredictionWriter
from viscy.trainer import VisCyTrainer
from viscy.transforms import NormalizeSampled

# %% [markdown]
"""
## Data and Model Paths

The dataset and model checkpoint files need to be downloaded before running this example.
"""

# %% [markdown] tags=[]
#
# <div class="alert alert-block alert-info">
#
# # Download the dataset and checkpoints
#
# - Download the neuromast test dataset and model checkpoint from here: <br>
# https://public.czbiohub.org/comp.micro/viscy
# - Update the `input_data_path` and `model_ckpt_path` variables with the path to the downloaded files.
# - Select a FOV (i.e 0/3/0).
# - Set an output path for the predictions.
#
# </div>
# %%
# TODO: modify the path to the downloaded dataset
input_data_path = "/20230803_fish2_60x_1_cropped_zyx_resampled_clipped_2.zarr"

# TODO: modify the path to the downloaded checkpoint
model_ckpt_path = "/epoch=44-step=1215.ckpt"

# TODO: modify the path
# Zarr store to save the predictions
output_path = "./test_neuromast_demo.zarr"

# TODO: Choose an FOV
# FOV of interest
fov = "0/3/0"

input_data_path = Path(input_data_path) / fov
# %%
# Create the VSNeuromast model

# Reduce the batch size if encountering out-of-memory errors
BATCH_SIZE = 2
# NOTE: Set the number of workers to 0 for Windows and macOS
# since multiprocessing only works with a
# `if __name__ == '__main__':` guard.
# On Linux, set it to the number of CPU cores to maximize performance.
NUM_WORKERS = 0
phase_channel_name = "Phase3D"

# %%[markdown]
"""
For this example we will use the following parameters:
### For more information on the VSNeuromast model:
See ``viscy.unet.networks.unext2`` ([source code](https://github.com/mehta-lab/VisCy/blob/main/viscy/unet/networks/unext2.py)) for configuration details.
"""
# %%
# Setup the data module.
data_module = HCSDataModule(
    data_path=input_data_path,
    source_channel=phase_channel_name,
    target_channel=["Membrane", "Nuclei"],
    z_window_size=21,
    split_ratio=0.8,
    batch_size=BATCH_SIZE,
    num_workers=NUM_WORKERS,
    architecture="UNeXt2",
    normalizations=[
        NormalizeSampled(
            [phase_channel_name],
            level="fov_statistics",
            subtrahend="median",
            divisor="iqr",
        )
    ],
)
data_module.prepare_data()
data_module.setup(stage="predict")
# %%
# Setup the model.
# Dictionary that specifies key parameters of the model.
config_VSNeuromast = {
    "in_channels": 1,
    "out_channels": 2,
    "in_stack_depth": 21,
    "backbone": "convnextv2_tiny",
    "stem_kernel_size": (7, 4, 4),
    "decoder_mode": "pixelshuffle",
    "head_expansion_ratio": 4,
    "head_pool": True,
}

model_VSNeuromast = VSUNet.load_from_checkpoint(
    model_ckpt_path, architecture="UNeXt2", model_config=config_VSNeuromast
)
model_VSNeuromast.eval()

# %%
# Setup the Trainer
trainer = VisCyTrainer(
    accelerator="gpu",
    callbacks=[HCSPredictionWriter(output_path)],
)

# Start the predictions
trainer.predict(
    model=model_VSNeuromast,
    datamodule=data_module,
    return_predictions=False,
)

# %%
# Open the output_zarr store and inspect the output
# Show the individual channels and the fused in a 1x3 plot
output_path = Path(output_path) / fov

# %%
# Open the predicted data
vs_store = open_ome_zarr(output_path, mode="r")
T, C, Z, Y, X = vs_store.data.shape
# Get a z-slice
z_slice = Z // 2  # NOTE: using the middle slice of the stack. Change as needed.
vs_nucleus = vs_store[0][0, 0, z_slice]  # (t,c,z,y,x)
vs_membrane = vs_store[0][0, 1, z_slice]  # (t,c,z,y,x)

# Open the experimental fluorescence
fluor_store = open_ome_zarr(input_data_path, mode="r")
# Get the 2D images
# NOTE: Channel indeces hardcoded for this dataset
fluor_nucleus = fluor_store[0][0, 1, z_slice]  # (t,c,z,y,x)
fluor_membrane = fluor_store[0][0, 2, z_slice]  # (t,c,z,y,x)

# Plot
plot_vs_n_fluor(vs_nucleus, vs_membrane, fluor_nucleus, fluor_membrane)

vs_store.close()
fluor_store.close()



from pathlib import Path

from iohub import open_ome_zarr
from monai.transforms import RandSpatialCropSamplesd
from pytest import mark

from viscy.data.hcs import HCSDataModule
from viscy.trainer import VisCyTrainer


@mark.parametrize("default_channels", [True, False])
def test_preprocess(small_hcs_dataset: Path, default_channels: bool):
    data_path = small_hcs_dataset
    if default_channels:
        channel_names = -1
    else:
        with open_ome_zarr(data_path) as dataset:
            channel_names = dataset.channel_names
    trainer = VisCyTrainer(accelerator="cpu")
    trainer.preprocess(data_path, channel_names=channel_names, num_workers=2)
    with open_ome_zarr(data_path) as dataset:
        channel_names = dataset.channel_names
        for channel in channel_names:
            assert "dataset_statistics" in dataset.zattrs["normalization"][channel]
        for _, fov in dataset.positions():
            norm_metadata = fov.zattrs["normalization"]
            for channel in channel_names:
                assert channel in norm_metadata
                assert "dataset_statistics" in norm_metadata[channel]
                assert "fov_statistics" in norm_metadata[channel]


@mark.parametrize("multi_sample_augmentation", [True, False])
def test_datamodule_setup_fit(preprocessed_hcs_dataset, multi_sample_augmentation):
    data_path = preprocessed_hcs_dataset
    z_window_size = 5
    channel_split = 2
    split_ratio = 0.8
    yx_patch_size = [128, 96]
    batch_size = 4
    with open_ome_zarr(data_path) as dataset:
        channel_names = dataset.channel_names
    if multi_sample_augmentation:
        transforms = [
            RandSpatialCropSamplesd(
                keys=channel_names,
                roi_size=[z_window_size, *yx_patch_size],
                num_samples=2,
            )
        ]
    else:
        transforms = []
    dm = HCSDataModule(
        data_path=data_path,
        source_channel=channel_names[:channel_split],
        target_channel=channel_names[channel_split:],
        z_window_size=z_window_size,
        batch_size=batch_size,
        num_workers=0,
        augmentations=transforms,
        target_2d=False,
        split_ratio=split_ratio,
        yx_patch_size=yx_patch_size,
    )
    dm.setup(stage="fit")
    for batch in dm.train_dataloader():
        assert batch["source"].shape == (
            batch_size,
            channel_split,
            z_window_size,
            *yx_patch_size,
        )
        assert batch["target"].shape == (
            batch_size,
            len(channel_names) - channel_split,
            z_window_size,
            *yx_patch_size,
        )


@mark.parametrize("z_window_size", [1, 5])
def test_datamodule_setup_predict(preprocessed_hcs_dataset, z_window_size):
    data_path = preprocessed_hcs_dataset
    channel_split = 2
    with open_ome_zarr(data_path) as dataset:
        channel_names = dataset.channel_names
        img = next(dataset.positions())[1][0]
        total_p = len(list(dataset.positions()))
    dm = HCSDataModule(
        data_path=data_path,
        source_channel=channel_names[:channel_split],
        target_channel=channel_names[channel_split:],
        z_window_size=z_window_size,
        target_2d=bool(z_window_size == 1),
        batch_size=2,
        num_workers=0,
    )
    dm.setup(stage="predict")
    dataset = dm.predict_dataset
    assert len(dataset) == total_p * 2 * (img.slices - z_window_size + 1)
    assert dataset[0]["source"].shape == (
        channel_split,
        z_window_size,
        img.height,
        img.width,
    )



import logging
from pathlib import Path
from typing import Any

from xarray import open_zarr

from viscy.representation.evaluation.dimensionality_reduction import compute_phate

_logger = logging.getLogger(__name__)


def update_phate_embeddings(
    dataset_path: Path,
    phate_kwargs: dict[str, Any],
) -> None:
    """
    Update PHATE embeddings in an existing dataset with new parameters.

    Parameters
    ----------
    dataset_path : Path
        Path to the zarr store containing embeddings
    phate_kwargs : dict
        New PHATE parameters to use for recomputing embeddings.
        Common parameters include:
        - n_components: int, number of dimensions (default: 2)
        - knn: int, number of nearest neighbors (default: 5)
        - decay: int, decay rate for kernel (default: 40)
        - n_jobs: int, number of jobs for parallel processing
        - t: int, number of diffusion steps
        - gamma: float, gamma parameter for kernel
    """
    # Load dataset
    dataset = open_zarr(dataset_path, mode="r+")

    # Compute new PHATE embeddings and update dataset
    _logger.info(f"Computing PHATE embeddings with parameters: {phate_kwargs}")
    compute_phate(dataset, update_dataset=True, **phate_kwargs)

    _logger.info(f"Updated PHATE embeddings in {dataset_path}")



# %% [markdown] tags=[]
# # Virtual staining of phase contrast images using VSCyto3D with and without augmentations
#
# Written by Eduardo Hirata-Miyasaki, Ziwen Liu, and Shalin Mehta, CZ Biohub San Francisco
#
# ## Overview
#
# This notebook demonstrates how to use the VSCyto3D model to virtually stain phase contrast images. The phase contrast images were not part of the training.
# We will use the VSCyto3D model to predict the nuclei and cell membrane channels from a phase contrast image with two models:
# - One model trained without augmentations
# - One model trained with augmentations
#

# %% tags=[]
# Imports
from pathlib import Path
from iohub import open_ome_zarr
from viscy.data.hcs import HCSDataModule

# Viscy classes for the trainer and model
from viscy.translation.engine import VSUNet
from viscy.trainer import VisCyTrainer
from viscy.transforms import NormalizeSampled
from lightning.pytorch import seed_everything

import matplotlib.pyplot as plt
import numpy as np
import torch

# seed random number generators for reproducibility.
seed_everything(42, workers=True)
# %%
# Paths to data and log directory
top_dir = Path(
    "~/data/vs_PhC"
).expanduser()  # If this fails, make sure this to point to your data directory in the shared mounting point inside /dlmbl/data

# Path to the training data
data_path = top_dir / "test/HEK_H2B_CAAX_PhC_40x_registered.zarr"

# %% [markdown] tags=[]
# ## Load OME-Zarr Dataset

# There should be 34 FOVs in the dataset.
#
# Each FOV consists of 3 channels of 2048x2048 images,
# saved in the [High-Content Screening (HCS) layout](https://ngff.openmicroscopy.org/latest/#hcs-layout)
# specified by the Open Microscopy Environment Next Generation File Format
# (OME-NGFF).
#
# The 3 channels correspond to the QPI, nuclei, and cell membrane. The nuclei were stained with DAPI and the cell membrane with Cellmask.
#
# - The layout on the disk is: `row/col/field/pyramid_level/timepoint/channel/z/y/x.`
# - These datasets only have 1 level in the pyramid (highest resolution) which is '0'.
# %%
# Open dataset and look at it's structure
dataset = open_ome_zarr(data_path)
dataset.print_tree()
# %%
row = 0
col = 3
field = "000000"  # TODO Change this for a different FOV

# NOTE: this dataset only has one level
pyaramid_level = 0

fov_path = f"{row}/{col}/{field}"
input_data_path = Path(data_path) / fov_path
image = dataset[fov_path][pyaramid_level].numpy()

n_channels = len(dataset.channel_names)
Z, Y, X = image.shape[-3:]
figure, axes = plt.subplots(1, n_channels, figsize=(9, 3))
title_names = ["PhC", "TXR", "Y5"]
for i in range(n_channels):
    for i in range(n_channels):
        channel_image = image[0, i, Z // 2]
        # Invert the phase contrast channel
        if i == 0:
            channel_image = channel_image * -1
        # Adjust contrast to 0.5th and 99.5th percentile of pixel values.
        p_low, p_high = np.percentile(channel_image, (0.5, 99.5))
        channel_image = np.clip(channel_image, p_low, p_high)
        axes[i].imshow(channel_image, cmap="gray")
        axes[i].axis("off")
        axes[i].set_title(title_names[i])
plt.tight_layout()

# %% [markdown] tags=[]
# ## Create the VSCyto3D model
# Here we will instantiate the `HCSDataModule` that reads the ome-zarr dataset and prepares the data for inference.
# %%
# Reduce the batch size if encountering out-of-memory errors
BATCH_SIZE = 5
# NOTE: Set the number of workers to 0 for Windows and macOS
# since multiprocessing only works with a
# `if __name__ == '__main__':` guard.
# On Linux, set it to the number of CPU cores to maximize performance.
NUM_WORKERS = 0
source_channel_name = "BF"

# %%[markdown]
"""
For this example we will use the following parameters:
### For more information on the VSCyto3D model:
See ``viscy.unet.networks.fcmae``
([source code](https://github.com/mehta-lab/VisCy/blob/6a3457ec8f43ecdc51b1760092f1a678ed73244d/viscy/unet/networks/unext2.py#L252))
for configuration details.
"""
# %%
# Setup the data module.
data_module = HCSDataModule(
    data_path=input_data_path,
    source_channel=source_channel_name,
    target_channel=["Nuclei", "Membrane"],
    z_window_size=5,
    split_ratio=0.8,
    batch_size=BATCH_SIZE,
    num_workers=NUM_WORKERS,
    architecture="UNeXt2",
    normalizations=[
        NormalizeSampled(
            [source_channel_name],
            level="fov_statistics",
            subtrahend="median",
            divisor="iqr",
        )
    ],
)
data_module.prepare_data()
data_module.setup(stage="predict")

# %% [markdown] tags=[]
# ## Setup the _VSCyto3D_ model with and without augmentations
# We will load the model checkpoints and run inference on the phase contrast image.abs
# The model that utilizes augmentations shows better performance in the prediction of the nuclei and cell membrane channels.
# The phase contrast images were not part of the training for the `VSCyto3D`` model.
# %%

# TODO: change if you want to use a different GPU
GPU_ID = 0

# TODO: point to the downloaded model checkpoints
no_augmentation_model_ckpt = (
    top_dir / "models/no_augmentations/best_epoch=30-step=6076.ckpt"
)
VSCyto3D_model_ckpt = top_dir / "models/epoch=48-step=18130.ckpt"

# Dictionary that specifies key parameters of the model.
config_VSCyto3D = {
    "in_channels": 1,
    "out_channels": 2,
    "in_stack_depth": 5,
    "backbone": "convnextv2_tiny",
    "stem_kernel_size": (5, 4, 4),
    "decoder_mode": "pixelshuffle",
    "head_expansion_ratio": 4,
    "head_pool": True,
}

# Model without augmentation
model_VSCyto3D_no_augmentation = VSUNet.load_from_checkpoint(
    no_augmentation_model_ckpt, architecture="UNeXt2", model_config=config_VSCyto3D
)
model_VSCyto3D_no_augmentation.eval()
# Model with augmentation
model_VSCyto3D_w_augmentation = VSUNet.load_from_checkpoint(
    VSCyto3D_model_ckpt, architecture="UNeXt2", model_config=config_VSCyto3D
)
model_VSCyto3D_w_augmentation.eval()

# Setup the Trainer
trainer = VisCyTrainer(accelerator="gpu", devices=[GPU_ID], precision="16-mixed")

n = 5
patch_size = 256
y_slice = slice(Y // 2 - patch_size * n // 2, Y // 2 + patch_size * n // 2)
x_slice = slice(X // 2 - patch_size * n // 2, X // 2 + patch_size * n // 2)

# Get the Phase Contrast channel
c_idx = dataset.channel_names.index(source_channel_name)
phase_image = image[0:1, c_idx : c_idx + 1, Z // 2 - 3 : Z // 2 + 3, y_slice, x_slice]
# Normalize the image
median = dataset[fov_path].zattrs["normalization"][source_channel_name][
    "fov_statistics"
]["median"]
iqr = dataset[fov_path].zattrs["normalization"][source_channel_name]["fov_statistics"][
    "iqr"
]
phase_image = ((phase_image - median) / iqr) * -1

# Load the image to device
device = model_VSCyto3D_no_augmentation.device
phase_image = torch.tensor(phase_image).to(device)

# Run inference on the given volume
with torch.inference_mode():  # turn off gradient computation.
    pred_no_augmentation = model_VSCyto3D_no_augmentation(phase_image)
    pred_w_augmentation = model_VSCyto3D_w_augmentation(phase_image)

pred_no_augmentation = pred_no_augmentation.cpu().detach().numpy()
pred_w_augmentation = pred_w_augmentation.cpu().detach().numpy()
phase_image = phase_image.cpu().detach().numpy()
clim_max = 30
clim_min = -20

# Plot the predicted images with model without augmentations
fig, ax = plt.subplots(2, 3, figsize=(12, 12))
ax[0, 0].imshow(phase_image[0, 0, 2, :, :], cmap="gray", vmin=clim_min, vmax=clim_max)
ax[0, 0].axis("off")
ax[0, 0].set_title("Phase Contrast")
for i in range(2):
    ax[0, i + 1].imshow(pred_no_augmentation[0, i, 2, :, :], cmap="gray")
    ax[0, i + 1].axis("off")
ax[0, 1].set_title("VS_Nuclei without augmentations")
ax[0, 2].set_title("VS_Membrane without augmentations")

# Plot the predicted images with VSCyto3D with augmentations
ax[1, 0].imshow(phase_image[0, 0, 2, :, :], cmap="gray", vmin=clim_min, vmax=clim_max)
ax[1, 0].axis("off")
ax[1, 0].set_title("Phase Contrast")
for i in range(2):
    ax[1, i + 1].imshow(
        pred_w_augmentation[0, i, 2, :, :],
        cmap="gray",
    )
    ax[1, i + 1].axis("off")
ax[1, 1].set_title("VS_Nuclei with augmentations")
ax[1, 2].set_title("VS_Membrane with augmentations")

plt.tight_layout()



import numpy as np
import scipy.ndimage as ndimage
from scipy.ndimage import binary_fill_holes
from skimage.filters import gaussian, laplace, threshold_otsu
from skimage.morphology import (
    ball,
    binary_dilation,
    binary_opening,
    disk,
    remove_small_objects,
)


def create_otsu_mask(input_image, sigma=0.6):
    """Create a binary mask using morphological operations
    :param np.array input_image: generate masks from this 3D image
    :param float sigma: Gaussian blur standard deviation,
        increase in value increases blur
    :return: volume mask of input_image, 3D np.array
    """

    input_sz = input_image.shape
    mid_slice_id = input_sz[0] // 2

    thresh = threshold_otsu(input_image[mid_slice_id, :, :])
    mask = input_image >= thresh

    return mask


def create_membrane_mask(input_image, str_elem_size=23, sigma=0.4, k_size=3, msize=120):
    """Create a binary mask using Laplacian of Gaussian (LOG) feature detection

    :param np.array input_image: generate masks from this image
    :param int str_elem_size: size of the laplacian filter
        used for contarst enhancement, odd number.
        Increase in value increases sensitivity of contrast enhancement
    :param float sigma: Gaussian blur standard deviation
    :param int k_size: disk/ball size for mask dilation,
        ball for 3D and disk for 2D data
    :param int msize: size of small objects removed to clean segmentation
    :return: mask of input_image, np.array
    """

    input_image_blur = gaussian(input_image, sigma=sigma)

    input_Lapl = laplace(input_image_blur, ksize=str_elem_size)

    thresh = threshold_otsu(input_Lapl)
    mask_bin = input_Lapl >= thresh

    if len(input_image.shape) == 2:
        str_elem = disk(k_size)
    else:
        str_elem = ball(k_size)

    mask_dilated = binary_dilation(mask_bin, str_elem)

    mask = remove_small_objects(mask_dilated, min_size=msize)

    return mask


def get_unimodal_threshold(input_image):
    """Determines optimal unimodal threshold

    https://users.cs.cf.ac.uk/Paul.Rosin/resources/papers/unimodal2.pdf
    https://www.mathworks.com/matlabcentral/fileexchange/45443-rosin-thresholding

    :param np.array input_image: generate mask for this image
    :return float best_threshold: optimal lower threshold for the foreground
     hist
    """

    hist_counts, bin_edges = np.histogram(
        input_image,
        bins=256,
        range=(input_image.min(), np.percentile(input_image, 99.5)),
    )
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

    # assuming that background has the max count
    max_idx = np.argmax(hist_counts)
    int_with_max_count = bin_centers[max_idx]
    p1 = [int_with_max_count, hist_counts[max_idx]]

    # find last non-empty bin
    pos_counts_idx = np.where(hist_counts > 0)[0]
    last_binedge = pos_counts_idx[-1]
    p2 = [bin_centers[last_binedge], hist_counts[last_binedge]]

    best_threshold = -np.inf
    max_dist = -np.inf
    for idx in range(max_idx, last_binedge, 1):
        x0 = bin_centers[idx]
        y0 = hist_counts[idx]
        a = [p1[0] - p2[0], p1[1] - p2[1]]
        b = [x0 - p2[0], y0 - p2[1]]
        cross_ab = a[0] * b[1] - b[0] * a[1]
        per_dist = np.linalg.norm(cross_ab) / np.linalg.norm(a)
        if per_dist > max_dist:
            best_threshold = x0
            max_dist = per_dist
    assert best_threshold > -np.inf, "Error in unimodal thresholding"
    return best_threshold


def create_unimodal_mask(input_image, str_elem_size=3, sigma=0.6):
    """
    Create a mask with unimodal thresholding and morphological operations.
    Unimodal thresholding seems to oversegment, erode it by a fraction

    :param np.array input_image: generate masks from this image
    :param int str_elem_size: size of the structuring element. typically 3, 5
    :param float sigma: gaussian blur standard deviation
    :return mask of input_image, np.array
    """

    input_image = gaussian(input_image, sigma=sigma)

    if np.min(input_image) == np.max(input_image):
        thr = np.unique(input_image)
    else:
        thr = get_unimodal_threshold(input_image)
    if len(input_image.shape) == 2:
        str_elem = disk(str_elem_size)
    else:
        str_elem = ball(str_elem_size)
    # remove small objects in mask
    mask = input_image >= thr
    mask = binary_opening(mask, str_elem)
    mask = binary_fill_holes(mask)
    return mask


def get_unet_border_weight_map(annotation, w0=10, sigma=5):
    """
    Return weight map for borders as specified in UNet paper
    :param annotation A 2D array of shape (image_height, image_width)
     contains annotation with each class labeled as an integer.
    :param w0 multiplier to the exponential distance loss
     default 10 as mentioned in UNet paper
    :param sigma standard deviation in the exponential distance term
     e^(-d1 + d2) ** 2 / 2 (sigma ^ 2)
     default 5 as mentioned in UNet paper
    :return weight mapt for borders as specified in UNet

    TODO: Calculate boundaries directly and calculate distance
    from boundary of cells to another
    Note: The below method only works for UNet Segmentation only
    """
    # if there is only one label, zero return the array as is
    if np.sum(annotation) == 0:
        return annotation

    # Masks could be saved as .npy bools, if so convert to uint8 and generate
    # labels from binary
    if annotation.dtype == bool:
        annotation = annotation.astype(np.uint8)
    assert annotation.dtype in [
        np.uint8,
        np.uint16,
    ], "Expected data type uint, it is {}".format(annotation.dtype)

    # cells instances for distance computation
    # 4 connected i.e default (cross-shaped)
    # structuring element to measure connectivy
    # If cells are 8 connected/touching they are labeled as one single object
    # Loss metric on such borders is not useful
    labeled_array, _ = ndimage.measurements.label(annotation)
    # class balance weights w_c(x)
    unique_values = np.unique(labeled_array).tolist()
    weight_map = [0] * len(unique_values)
    for index, unique_value in enumerate(unique_values):
        mask = np.zeros((annotation.shape[0], annotation.shape[1]), dtype=np.float64)
        mask[annotation == unique_value] = 1
        weight_map[index] = 1 / mask.sum()

    # this normalization is important - foreground pixels must have weight 1
    weight_map = [i / max(weight_map) for i in weight_map]

    wc = np.zeros((annotation.shape[0], annotation.shape[1]), dtype=np.float64)
    for index, unique_value in enumerate(unique_values):
        wc[annotation == unique_value] = weight_map[index]

    # cells distance map
    border_loss_map = np.zeros(
        (annotation.shape[0], annotation.shape[1]), dtype=np.float64
    )
    distance_maps = np.zeros(
        (annotation.shape[0], annotation.shape[1], np.max(labeled_array)),
        dtype=np.float64,
    )

    if np.max(labeled_array) >= 2:
        for index in range(np.max(labeled_array)):
            mask = np.ones_like(labeled_array)
            mask[labeled_array == index + 1] = 0
            distance_maps[:, :, index] = ndimage.distance_transform_edt(mask)
    distance_maps = np.sort(distance_maps, 2)
    d1 = distance_maps[:, :, 0]
    d2 = distance_maps[:, :, 1]
    border_loss_map = w0 * np.exp((-1 * (d1 + d2) ** 2) / (2 * (sigma**2)))

    zero_label = np.zeros((annotation.shape[0], annotation.shape[1]), dtype=np.float64)
    zero_label[labeled_array == 0] = 1
    border_loss_map = np.multiply(border_loss_map, zero_label)

    # unet weight map mask
    weight_map_mask = wc + border_loss_map

    return weight_map_mask



import logging
from pathlib import Path
from typing import Literal, Sequence

import numpy as np
import pandas as pd
import tensorstore as ts
import torch
from iohub.ngff import ImageArray, Position, open_ome_zarr
from monai.data import ThreadDataLoader
from monai.data.utils import collate_meta_tensor
from monai.transforms import Compose, MapTransform, ToDeviced
from torch import Tensor
from torch.utils.data import Dataset

from viscy.data.hcs import HCSDataModule, _read_norm_meta
from viscy.data.select import _filter_fovs, _filter_wells
from viscy.data.typing import DictTransform, NormMeta, TripletSample

_logger = logging.getLogger("lightning.pytorch")

INDEX_COLUMNS = [
    "fov_name",
    "track_id",
    "t",
    "id",
    "parent_track_id",
    "parent_id",
    "z",
    "y",
    "x",
]


def _scatter_channels(
    channel_names: list[str], patch: Tensor, norm_meta: NormMeta | None
) -> dict[str, Tensor | NormMeta] | dict[str, Tensor]:
    channels = {
        name: patch[:, c : c + 1]
        for name, c in zip(channel_names, range(patch.shape[1]))
    }
    if norm_meta is not None:
        channels["norm_meta"] = collate_meta_tensor(norm_meta)
    return channels


def _gather_channels(
    patch_channels: list[dict[str, Tensor | NormMeta]],
) -> list[Tensor]:
    samples = []
    for sample in patch_channels:
        sample.pop("norm_meta", None)
        samples.append(torch.cat(list(sample.values()), dim=0))
    return samples


def _transform_channel_wise(
    transform: DictTransform,
    channel_names: list[str],
    patch: Tensor,
    norm_meta: NormMeta | None,
) -> list[Tensor]:
    scattered_channels = _scatter_channels(channel_names, patch, norm_meta)
    transformed_channels = transform(scattered_channels)
    return _gather_channels(transformed_channels)


class TripletDataset(Dataset):
    def __init__(
        self,
        positions: list[Position],
        tracks_tables: list[pd.DataFrame],
        channel_names: list[str],
        initial_yx_patch_size: tuple[int, int],
        z_range: slice,
        anchor_transform: DictTransform | None = None,
        positive_transform: DictTransform | None = None,
        negative_transform: DictTransform | None = None,
        fit: bool = True,
        predict_cells: bool = False,
        include_fov_names: list[str] | None = None,
        include_track_ids: list[int] | None = None,
        time_interval: Literal["any"] | int = "any",
        return_negative: bool = True,
    ) -> None:
        """Dataset for triplet sampling of cells based on tracking.

        Parameters
        ----------
        positions : list[Position]
            OME-Zarr images with consistent channel order
        tracks_tables : list[pd.DataFrame]
            Data frames containing ultrack results
        channel_names : list[str]
            Input channel names
        initial_yx_patch_size : tuple[int, int]
            YX size of the initially sampled image patch before augmentation
        z_range : slice
            Range of Z-slices
        anchor_transform : DictTransform | None, optional
            Transforms applied to the anchor sample, by default None
        positive_transform : DictTransform | None, optional
            Transforms applied to the positve sample, by default None
        negative_transform : DictTransform | None, optional
            Transforms applied to the negative sample, by default None
        fit : bool, optional
            Fitting mode in which the full triplet will be sampled,
            only sample anchor if ``False``, by default True
        predict_cells : bool, optional
            Only predict on selected cells, by default False
        include_fov_names : list[str] | None, optional
            Only predict on selected FOVs, by default None
        include_track_ids : list[int] | None, optional
            Only predict on selected track IDs, by default None
        time_interval : Literal["any"] | int, optional
            Future time interval to sample positive and anchor from,
            by default "any"
            (sample negative from another track any time point
            and use the augmented anchor patch as positive)
        return_negative : bool, optional
            Whether to return the negative sample during the fit stage
            (can be set to False when using a loss function like NT-Xent),
            by default True
        """
        self.positions = positions
        self.channel_names = channel_names
        self.channel_indices = [
            positions[0].get_channel_index(ch) for ch in channel_names
        ]
        self.z_range = z_range
        self.anchor_transform = anchor_transform
        self.positive_transform = positive_transform
        self.negative_transform = negative_transform
        self.fit = fit
        self.yx_patch_size = initial_yx_patch_size
        self.predict_cells = predict_cells
        self.include_fov_names = include_fov_names or []
        self.include_track_ids = include_track_ids or []
        self.time_interval = time_interval
        self.tracks = self._filter_tracks(tracks_tables)
        self.tracks = (
            self._specific_cells(self.tracks) if self.predict_cells else self.tracks
        )
        self.valid_anchors = self._filter_anchors(self.tracks)
        self.return_negative = return_negative

    def _filter_tracks(self, tracks_tables: list[pd.DataFrame]) -> pd.DataFrame:
        """Exclude tracks that are too close to the border or do not have the next time point.

        Parameters
        ----------
        tracks_tables : list[pd.DataFrame]
            List of tracks_tables returned by TripletDataModule._align_tracks_tables_with_positions

        Returns
        -------
        pd.DataFrame
            Filtered tracks table
        """
        filtered_tracks = []
        y_exclude, x_exclude = (self.yx_patch_size[0] // 2, self.yx_patch_size[1] // 2)
        for pos, tracks in zip(self.positions, tracks_tables, strict=True):
            tracks["position"] = [pos] * len(tracks)
            tracks["fov_name"] = pos.zgroup.name
            tracks["global_track_id"] = tracks["fov_name"].str.cat(
                tracks["track_id"].astype(str), sep="_"
            )
            image: ImageArray = pos["0"]
            if self.z_range.stop > image.slices:
                raise ValueError(
                    f"Z range {self.z_range} exceeds image with Z={image.slices}"
                )
            y_range = (y_exclude, image.height - y_exclude)
            x_range = (x_exclude, image.width - x_exclude)
            # FIXME: Check if future time points are available after interval
            filtered_tracks.append(
                tracks[
                    tracks["y"].between(*y_range, inclusive="neither")
                    & tracks["x"].between(*x_range, inclusive="neither")
                ]
            )
        return pd.concat(filtered_tracks).reset_index(drop=True)

    def _filter_anchors(self, tracks: pd.DataFrame) -> pd.DataFrame:
        """Ensure that anchors have the next time point after a time interval."""
        if self.time_interval == "any" or not self.fit:
            return tracks
        return pd.concat(
            [
                track[(track["t"] + self.time_interval).isin(track["t"])]
                for (_, track) in tracks.groupby("global_track_id")
            ]
        )

    def _specific_cells(self, tracks: pd.DataFrame) -> pd.DataFrame:
        specific_tracks = pd.DataFrame()
        print(self.include_fov_names)
        print(self.include_track_ids)
        for fov_name, track_id in zip(self.include_fov_names, self.include_track_ids):
            filtered_tracks = tracks[
                (tracks["fov_name"] == fov_name) & (tracks["track_id"] == track_id)
            ]
            specific_tracks = pd.concat([specific_tracks, filtered_tracks])
        return specific_tracks.reset_index(drop=True)

    def __len__(self) -> int:
        return len(self.valid_anchors)

    def _sample_positives(self, anchor_rows: pd.DataFrame) -> pd.DataFrame:
        """Select a positive sample from the same track in the next time point."""
        query = anchor_rows[["global_track_id", "t"]].copy()
        query["t"] += self.time_interval
        return query.merge(self.tracks, on=["global_track_id", "t"], how="inner")

    def _sample_negative(self, anchor_row: pd.Series) -> pd.Series:
        """Select a negative sample from a different track in the next time point
        if an interval is specified, otherwise from any random time point."""
        if self.time_interval == "any":
            tracks = self.tracks
        else:
            tracks = self.tracks[
                self.tracks["t"] == anchor_row["t"] + self.time_interval
            ]
        candidates: pd.DataFrame = tracks[
            (tracks["global_track_id"] != anchor_row["global_track_id"])
        ]
        # NOTE: Random sampling
        # this is to avoid combinatorial length growth at fitting time
        # since each cell can pair with any other cell
        # (3e4 instances will make 1e9 pairs)
        # reproducibility relies on setting a global seed for numpy
        return candidates.sample(n=1).iloc[0]

    def _sample_negatives(self, anchor_rows: pd.DataFrame) -> pd.DataFrame:
        return pd.concat(
            [self._sample_negative(row) for _, row in anchor_rows.iterrows()],
            axis=1,
        )

    def _slice_patch(
        self, track_row: pd.Series
    ) -> tuple[ts.TensorStore, NormMeta | None]:
        position: Position = track_row["position"]
        image = position["0"].tensorstore()
        time = track_row["t"]
        y_center = track_row["y"]
        x_center = track_row["x"]
        y_half, x_half = (d // 2 for d in self.yx_patch_size)
        patch = image.oindex[
            time,
            [int(i) for i in self.channel_indices],
            self.z_range,
            slice(y_center - y_half, y_center + y_half),
            slice(x_center - x_half, x_center + x_half),
        ]
        return patch, _read_norm_meta(position)

    def _slice_patches(self, track_rows: pd.DataFrame):
        patches = []
        norms = []
        with ts.Batch() as batch:
            for _, row in track_rows.iterrows():
                patch, norm = self._slice_patch(row)
                patches.append(patch.read(batch=batch))
                norms.append(norm)
        results = [p.result() for p in patches]
        return torch.from_numpy(np.stack(results, axis=0)), norms

    def __getitems__(self, indices: list[int]) -> list[TripletSample]:
        anchor_rows = self.valid_anchors.iloc[indices]
        anchor_patches, anchor_norms = self._slice_patches(anchor_rows)
        if self.fit:
            if self.time_interval == "any":
                positive_patches = anchor_patches.clone()
                positive_norms = anchor_norms
            else:
                positive_rows = self._sample_positives(anchor_rows)
                positive_patches, positive_norms = self._slice_patches(positive_rows)
            if self.positive_transform:
                positive_patches = _transform_channel_wise(
                    transform=self.positive_transform,
                    channel_names=self.channel_names,
                    patch=positive_patches,
                    norm_meta=positive_norms,
                )
            if self.return_negative:
                negative_rows = self._sample_negatives(anchor_rows)
                negative_patches, negative_norms = self._slice_patches(negative_rows)
                if self.negative_transform:
                    negative_patches = _transform_channel_wise(
                        transform=self.negative_transform,
                        channel_names=self.channel_names,
                        patch=negative_patches,
                        norm_meta=negative_norms,
                    )
        if self.anchor_transform:
            anchor_patches = _transform_channel_wise(
                transform=self.anchor_transform,
                channel_names=self.channel_names,
                patch=anchor_patches,
                norm_meta=anchor_norms,
            )
        samples: list[TripletSample] = [
            {"anchor": anchor_patch} for anchor_patch in anchor_patches
        ]
        if self.fit:
            for sample, positive_patch in zip(samples, positive_patches):
                sample["positive"] = positive_patch
            if self.return_negative:
                for sample, negative_patch in zip(samples, negative_patches):
                    sample["negative"] = negative_patch
        else:
            for sample, (_, anchor_row) in zip(samples, anchor_rows.iterrows()):
                # For new predictions, ensure all INDEX_COLUMNS are included
                index_dict = {}
                for col in INDEX_COLUMNS:
                    if col in anchor_row.index:
                        index_dict[col] = anchor_row[col]
                    elif col not in ["y", "x", "z"]:
                        # Skip y and x for legacy data - they weren't part of INDEX_COLUMNS
                        raise KeyError(f"Required column '{col}' not found in data")
                sample["index"] = index_dict
        return samples


class TripletDataModule(HCSDataModule):
    def __init__(
        self,
        data_path: str,
        tracks_path: str,
        source_channel: str | Sequence[str],
        z_range: tuple[int, int],
        initial_yx_patch_size: tuple[int, int] = (512, 512),
        final_yx_patch_size: tuple[int, int] = (224, 224),
        split_ratio: float = 0.8,
        batch_size: int = 16,
        num_workers: int = 8,
        normalizations: list[MapTransform] = [],
        augmentations: list[MapTransform] = [],
        caching: bool = False,
        fit_include_wells: list[str] | None = None,
        fit_exclude_fovs: list[str] | None = None,
        predict_cells: bool = False,
        include_fov_names: list[str] | None = None,
        include_track_ids: list[int] | None = None,
        time_interval: Literal["any"] | int = "any",
        return_negative: bool = True,
        persistent_workers: bool = False,
        prefetch_factor: int | None = None,
        pin_memory: bool = False,
        z_window_size: int | None = None,
    ):
        """Lightning data module for triplet sampling of patches.

        Parameters
        ----------
        data_path : str
            Image dataset path
        tracks_path : str
            Tracks labels dataset path
        source_channel : str | Sequence[str]
            List of input channel names
        z_range : tuple[int, int]
            Range of valid z-slices
        initial_yx_patch_size : tuple[int, int], optional
            XY size of the initially sampled image patch, by default (512, 512)
        final_yx_patch_size : tuple[int, int], optional
            Output patch size, by default (224, 224)
        split_ratio : float, optional
            Ratio of training samples, by default 0.8
        batch_size : int, optional
            Batch size, by default 16
        num_workers : int, optional
            Number of data-loading workers, by default 8
        normalizations : list[MapTransform], optional
            Normalization transforms, by default []
        augmentations : list[MapTransform], optional
            Augmentation transforms, by default []
        caching : bool, optional
            Whether to cache the dataset, by default False
        fit_include_wells : list[str], optional
            Only include these wells for fitting, by default None
        fit_exclude_fovs : list[str], optional
            Exclude these FOVs for fitting, by default None
        predict_cells : bool, optional
            Only predict for selected cells, by default False
        include_fov_names : list[str] | None, optional
            Only predict for selected FOVs, by default None
        include_track_ids : list[int] | None, optional
            Only predict for selected tracks, by default None
        time_interval : Literal["any"] | int, optional
            Future time interval to sample positive and anchor from,
            "any" means sampling negative from another track any time point
            and using the augmented anchor patch as positive), by default "any"
        return_negative : bool, optional
            Whether to return the negative sample during the fit stage
            (can be set to False when using a loss function like NT-Xent),
            by default True
        persistent_workers : bool, optional
            Whether to keep worker processes alive between iterations, by default False
        prefetch_factor : int | None, optional
            Number of batches loaded in advance by each worker, by default None
        pin_memory : bool, optional
            Whether to pin memory in CPU for faster GPU transfer, by default False
        z_window_size : int, optional
            Size of the final Z window, by default None (inferred from z_range)
        """
        super().__init__(
            data_path=data_path,
            source_channel=source_channel,
            target_channel=[],
            z_window_size=z_window_size or z_range[1] - z_range[0],
            split_ratio=split_ratio,
            batch_size=batch_size,
            num_workers=num_workers,
            target_2d=False,
            yx_patch_size=final_yx_patch_size,
            normalizations=normalizations,
            augmentations=augmentations,
            caching=caching,
            persistent_workers=persistent_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=pin_memory,
        )
        self.z_range = slice(*z_range)
        self.tracks_path = Path(tracks_path)
        self.initial_yx_patch_size = initial_yx_patch_size
        self._include_wells = fit_include_wells
        self._exclude_fovs = fit_exclude_fovs
        self.predict_cells = predict_cells
        self.include_fov_names = include_fov_names
        self.include_track_ids = include_track_ids
        self.time_interval = time_interval
        self.return_negative = return_negative

    def _align_tracks_tables_with_positions(
        self,
    ) -> tuple[list[Position], list[pd.DataFrame]]:
        """Parse positions in ome-zarr store containing tracking information
        and assemble tracks tables for each position.

        Returns
        -------
        tuple[list[Position], list[pd.DataFrame]]
            List of positions and list of tracks tables for each position
        """
        positions = []
        tracks_tables = []
        images_plate = open_ome_zarr(self.data_path)
        for well in _filter_wells(images_plate, include_wells=self._include_wells):
            for fov in _filter_fovs(well, exclude_fovs=self._exclude_fovs):
                positions.append(fov)
                tracks_df = pd.read_csv(
                    next((self.tracks_path / fov.zgroup.name.strip("/")).glob("*.csv"))
                ).astype(int)
                tracks_tables.append(tracks_df)

        return positions, tracks_tables

    @property
    def _base_dataset_settings(self) -> dict:
        return {
            "channel_names": self.source_channel,
            "z_range": self.z_range,
            "time_interval": self.time_interval,
        }

    def _update_to_device_transform(self):
        "Make sure that GPU transforms are set to the current device."
        for transform in self.normalizations + self.augmentations:
            if isinstance(transform, ToDeviced):
                transform.converter.device = torch.device(
                    f"cuda:{torch.cuda.current_device()}"
                )

    def _setup_fit(self, dataset_settings: dict):
        self._update_to_device_transform()
        augment_transform, no_aug_transform = self._fit_transform()
        positions, tracks_tables = self._align_tracks_tables_with_positions()
        shuffled_indices = self._set_fit_global_state(len(positions))
        positions = [positions[i] for i in shuffled_indices]
        tracks_tables = [tracks_tables[i] for i in shuffled_indices]

        num_train_fovs = int(len(positions) * self.split_ratio)
        train_positions = positions[:num_train_fovs]
        val_positions = positions[num_train_fovs:]
        train_tracks_tables = tracks_tables[:num_train_fovs]
        val_tracks_tables = tracks_tables[num_train_fovs:]
        _logger.debug(f"Number of training FOVs: {len(train_positions)}")
        _logger.debug(f"Number of validation FOVs: {len(val_positions)}")
        anchor_transform = (
            no_aug_transform
            if (self.time_interval == "any" or self.time_interval == 0)
            else augment_transform
        )
        self.train_dataset = TripletDataset(
            positions=train_positions,
            tracks_tables=train_tracks_tables,
            initial_yx_patch_size=self.initial_yx_patch_size,
            anchor_transform=anchor_transform,
            positive_transform=augment_transform,
            negative_transform=augment_transform,
            fit=True,
            return_negative=self.return_negative,
            **dataset_settings,
        )

        self.val_dataset = TripletDataset(
            positions=val_positions,
            tracks_tables=val_tracks_tables,
            initial_yx_patch_size=self.initial_yx_patch_size,
            anchor_transform=anchor_transform,
            positive_transform=augment_transform,
            negative_transform=augment_transform,
            fit=True,
            return_negative=self.return_negative,
            **dataset_settings,
        )

    def _setup_predict(self, dataset_settings: dict):
        self._set_predict_global_state()
        positions, tracks_tables = self._align_tracks_tables_with_positions()
        self.predict_dataset = TripletDataset(
            positions=positions,
            tracks_tables=tracks_tables,
            initial_yx_patch_size=self.initial_yx_patch_size,
            anchor_transform=Compose(self.normalizations),
            fit=False,
            predict_cells=self.predict_cells,
            include_fov_names=self.include_fov_names,
            include_track_ids=self.include_track_ids,
            **dataset_settings,
        )

    def _setup_test(self, *args, **kwargs):
        raise NotImplementedError("Self-supervised model does not support testing")

    def train_dataloader(self):
        return ThreadDataLoader(
            self.train_dataset,
            use_thread_workers=True,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
            prefetch_factor=self.prefetch_factor if self.num_workers else None,
            persistent_workers=self.persistent_workers,
            drop_last=True,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self):
        return ThreadDataLoader(
            self.val_dataset,
            use_thread_workers=True,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            prefetch_factor=self.prefetch_factor if self.num_workers else None,
            persistent_workers=self.persistent_workers,
            drop_last=False,
            pin_memory=self.pin_memory,
        )

    def predict_dataloader(self):
        return ThreadDataLoader(
            self.predict_dataset,
            use_thread_workers=True,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            prefetch_factor=self.prefetch_factor if self.num_workers else None,
            persistent_workers=self.persistent_workers,
            drop_last=False,
            pin_memory=self.pin_memory,
        )



# %% [markdown]
# # Demo: Comparing DynaCLR vs ImageNet Embeddings for Cell Infection Analysis
#
# This tutorial demonstrates how to:
# 1. Use ImageNet pre-trained features for analyzing cell infection
# 2. Compare with DynaCLR learned features
# 3. Visualize the differences between approaches

# %% [markdown]
# ## Setup and Imports

# %%
from pathlib import Path

import numpy as np
import pandas as pd
from skimage.exposure import rescale_intensity

from utils import (
    create_combined_visualization,
)
from viscy.data.triplet import TripletDataModule
from viscy.representation.embedding_writer import read_embedding_dataset

# %% [markdown]
# ## Set Data Paths
#
# The data, tracks, annotations and precomputed embeddings can be downloaded from [here]()
#
# ## Note:
#
# Alternatively, you can run the CLI to compute the features yourself by following the instructions in the [README.md](./README.md)

# %%
# TODO: Update the paths to the downloaded data
# Point to the *.zarr files
download_root = Path.home() / "data/dynaclr/demo"
input_data_path = (
    download_root / "registered_test.zarr"
)  # Replace with path to registered_test.zarr
tracks_path = download_root / "track_test.zarr"  # Replace with path to  track_test.zarr
ann_path = (
    download_root / "extracted_inf_state.csv"
)  # Replace with path to extracted_inf_state.csv

# TODO: Update the path to the DynaCLR and ImageNet features
# Point to the precomputed embeddings
dynaclr_features_path = (
    download_root / "precomputed_embeddings/infection_160patch_94ckpt_rev6_dynaclr.zarr"
)
imagenet_features_path = (
    download_root
    / "precomputed_embeddings/20240204_A549_DENV_ZIKV_sensor_only_imagenet.zarr"
)

# %% [markdown]
# ## Load the embeddings and annotations
# Load the embeddings you downloaded and append the human annotations to the dataframe

# %%
# Load the embeddings
dynaclr_embeddings = read_embedding_dataset(dynaclr_features_path)
imagenet_embeddings = read_embedding_dataset(imagenet_features_path)

dynaclr_features_df = dynaclr_embeddings["sample"].to_dataframe().reset_index(drop=True)
imagenet_features_df = (
    imagenet_embeddings["sample"].to_dataframe().reset_index(drop=True)
)

# Load the annotations and create a dataframe with the infection state
annotation = pd.read_csv(ann_path)
annotation["fov_name"] = "/" + annotation["fov_name"]

imagenet_features_df["infection"] = float("nan")

for index, row in annotation.iterrows():
    mask = (
        (imagenet_features_df["fov_name"] == row["fov_name"])
        & (imagenet_features_df["track_id"] == row["track_id"])
        & (imagenet_features_df["t"] == row["t"])
    )
    imagenet_features_df.loc[mask, "infection"] = row["infection_state"]
    mask = (
        (dynaclr_features_df["fov_name"] == row["fov_name"])
        & (dynaclr_features_df["track_id"] == row["track_id"])
        & (dynaclr_features_df["t"] == row["t"])
    )
    dynaclr_features_df.loc[mask, "infection"] = row["infection_state"]

# Filter out rows with infection state 0
imagenet_features_df = imagenet_features_df[imagenet_features_df["infection"] != 0]
dynaclr_features_df = dynaclr_features_df[dynaclr_features_df["infection"] != 0]

# %% [markdown]
# ## Choose a representative track for visualization

# %%
# NOTE: We have chosen these tracks to be representative of the data. Feel free to open the dataset and select other tracks
fov_name_mock = "/A/3/9"
track_id_mock = [19]
fov_name_inf = "/B/4/9"
track_id_inf = [42]

# Default parameters for the test dataset
z_range = (24, 29)
yx_patch_size = (160, 160)

channels_to_display = ["Phase3D", "RFP"]
fov_name_mock_list = [fov_name_mock] * len(track_id_mock)
fov_name_inf_list = [fov_name_inf] * len(track_id_inf)

conditions_to_compare = {
    "uninfected": {
        "fov_name_list": fov_name_mock_list,
        "track_id_list": track_id_mock,
    },
    "infected": {
        "fov_name_list": fov_name_inf_list,
        "track_id_list": track_id_inf,
    },
}

print("Caching sample images...")
image_cache = {}
for condition, condition_data in conditions_to_compare.items():
    dm = TripletDataModule(
        data_path=input_data_path,
        tracks_path=tracks_path,
        source_channel=channels_to_display,
        z_range=z_range,
        initial_yx_patch_size=yx_patch_size,
        final_yx_patch_size=yx_patch_size,
        include_fov_names=condition_data["fov_name_list"]
        * len(condition_data["track_id_list"]),
        include_track_ids=condition_data["track_id_list"],
        predict_cells=True,
        batch_size=1,
    )
    dm.setup("predict")

    condition_key = f"{condition}_cache"
    image_cache[condition_key] = {
        "fov_name": None,
        "track_id": None,
        "images_by_timepoint": {},
    }
    for i, patch in enumerate(dm.predict_dataloader()):
        fov_name = patch["index"]["fov_name"][0]
        track_id = patch["index"]["track_id"][0]
        images = patch["anchor"].numpy()[0]
        t = int(patch["index"]["t"][0])

        if image_cache[condition_key]["fov_name"] is None:
            image_cache[condition_key]["fov_name"] = fov_name
            image_cache[condition_key]["track_id"] = track_id

        z_idx = images.shape[1] // 2
        C, Z, Y, X = images.shape
        image_out = np.zeros((C, 1, Y, X), dtype=np.float32)
        # NOTE: here we are using the default percentile range for the RFP channel, change if using different channels or this threshold does not work
        for c_idx, channel in enumerate(channels_to_display):
            if channel in ["Phase3D", "DIC", "BF"]:
                image_out[c_idx] = images[c_idx, z_idx]
                image_out[c_idx] = (
                    image_out[c_idx] - image_out[c_idx].mean()
                ) / image_out[c_idx].std()
                image_out[c_idx] = rescale_intensity(image_out[c_idx], out_range=(0, 1))
            else:
                image_out[c_idx] = np.max(images[c_idx], axis=0)
                lower, upper = np.percentile(image_out[c_idx], (50, 99))
                image_out[c_idx] = (image_out[c_idx] - lower) / (upper - lower)
                image_out[c_idx] = rescale_intensity(image_out[c_idx], out_range=(0, 1))

        image_cache[condition_key]["images_by_timepoint"][t] = image_out

    print(
        f"Cached {condition_key} with {len(image_cache[condition_key]['images_by_timepoint'])} timepoints"
    )

# %%
print("Creating Cell Images and PHATE Embeddings Visualization...")
create_combined_visualization(
    image_cache,
    imagenet_features_df,
    dynaclr_features_df,
    highlight_tracks={
        1: [(fov_name_mock, track_id_mock[0])],  # Uninfected tracks
        2: [(fov_name_inf, track_id_inf[0])],  # Infected tracks
    },
    subplot_titles=[
        "Uninfected Phase",
        "Uninfected Viral Sensor",
        "Infected Phase",
        "Infected Viral Sensor",
    ],
    condition_keys=["uninfected_cache", "infected_cache"],
    channel_colormaps=["gray", "magma"],
    category_colors={1: "cornflowerblue", 2: "salmon"},
    highlight_colors={1: "blue", 2: "red"},
    category_labels={1: "Uninfected", 2: "Infected"},
    plot_size_xy=(1200, 600),
    title_location="top",
)

# Save the visualization as an interactive HTML file
fig = create_combined_visualization(
    image_cache,
    imagenet_features_df,
    dynaclr_features_df,
    highlight_tracks={
        1: [(fov_name_mock, track_id_mock[0])],  # Uninfected tracks
        2: [(fov_name_inf, track_id_inf[0])],  # Infected tracks
    },
    subplot_titles=[
        "Uninfected Phase",
        "Uninfected Viral Sensor",
        "Infected Phase",
        "Infected Viral Sensor",
    ],
    condition_keys=["uninfected_cache", "infected_cache"],
    channel_colormaps=["gray", "magma"],
    category_colors={1: "cornflowerblue", 2: "salmon"},
    highlight_colors={1: "blue", 2: "red"},
    category_labels={1: "Uninfected", 2: "Infected"},
    plot_size_xy=(1200, 600),
    title_location="top",
)

# Create output directory if it doesn't exist
output_dir = Path("output")
output_dir.mkdir(exist_ok=True)

# Save the interactive visualization
output_path = output_dir / "cell_infection_visualization.html"
fig.write_html(str(output_path))
print(f"Saved interactive visualization to: {output_path}")

# %% [markdown]
# ## Conclusion
#
# Time-aware sampling improved temporal continutiy and dynamic range of embeddings.
# These improvements can be seen in the PHATE projections of DynaCLR.
# The embeddings show smoother and higher dynamic range.
#



# %%
import lightning.pytorch as pl
from applications.infection_classification.classify_infection_2D import (
    SemanticSegUNet2D,
)
from pytorch_lightning.loggers import TensorBoardLogger

from viscy.data.hcs import HCSDataModule
from viscy.transforms import NormalizeSampled

# %% test the model on the test set
test_datapath = "/hpc/projects/intracellular_dashboard/viral-sensor/2024_04_25_BJ5a_DENV_TimeCourse/4-human_annotation/test_data.zarr"

data_module = HCSDataModule(
    data_path=test_datapath,
    source_channel=["TXR_Density3D", "Phase3D"],
    target_channel=["Inf_mask"],
    split_ratio=0.7,
    z_window_size=1,
    architecture="2D",
    num_workers=1,
    batch_size=1,
    normalizations=[
        NormalizeSampled(
            keys=["TXR_Density3D", "Phase3D"],
            level="fov_statistics",
            subtrahend="median",
            divisor="iqr",
        )
    ],
)

data_module.setup(stage="test")

# %% create trainer and input

logger = TensorBoardLogger(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_04_25_BJ5a_DENV_TimeCourse/5-infection_classifier/0-model_training/",
    name="logs",
)

trainer = pl.Trainer(
    logger=logger,
    default_root_dir="/hpc/projects/intracellular_dashboard/viral-sensor/2024_04_25_BJ5a_DENV_TimeCourse/5-infection_classifier/0-model_training/logs",
    log_every_n_steps=1,
    devices=1,  # Set the number of GPUs to use. This avoids run-time exception from distributed training when the node has multiple GPUs
)

model = SemanticSegUNet2D(
    in_channels=2,
    out_channels=3,
    ckpt_path="/hpc/projects/intracellular_dashboard/viral-sensor/2024_04_25_BJ5a_DENV_TimeCourse/5-infection_classifier/0-model_training/logs/checkpoint_epoch=206.ckpt",
)

# %% test the model

trainer.test(model=model, datamodule=data_module)

# %%



import nose.tools
import numpy as np
from skimage import draw
from skimage.filters import gaussian

from viscy.utils.masks import (
    create_unimodal_mask,
    get_unet_border_weight_map,
    get_unimodal_threshold,
)

uni_thr_tst_image = np.zeros((31, 31))
uni_thr_tst_image[5:10, 8:16] = 127
uni_thr_tst_image[11:21, 2:12] = 97
uni_thr_tst_image[8:12, 3:7] = 31
uni_thr_tst_image[17:29, 17:29] = 61
uni_thr_tst_image[3:14, 17:29] = 47


def test_get_unimodal_threshold():
    input_image = gaussian(uni_thr_tst_image, 1)
    best_thr = get_unimodal_threshold(input_image)
    nose.tools.assert_equal(np.floor(best_thr), 3.0)


def test_unimodal_thresholding():
    input_image = gaussian(uni_thr_tst_image, 1)
    mask = create_unimodal_mask(input_image, str_elem_size=0)
    nose.tools.assert_equal(input_image.shape, mask.shape)
    nose.tools.assert_true(mask.dtype, bool)
    # Check that mask is somewhat close to simple thresholding
    thresh_im = input_image > 3.04
    nose.tools.assert_true(
        np.abs(np.mean(mask) - np.mean(thresh_im)) < 0.1,
    )


def test_get_unet_border_weight_map():
    # Creating a test image with 3 circles
    # 2 close to each other and one far away
    radius = 10
    params = [(20, 16, radius), (44, 16, radius), (47, 47, radius)]
    mask = np.zeros((64, 64), dtype=np.uint8)
    for i, (cx, cy, radius) in enumerate(params):
        rr, cc = draw.disk((cx, cy), radius)
        mask[rr, cc] = i + 1

    weight_map = get_unet_border_weight_map(mask)

    max_weight_map = np.max(weight_map)
    # weight map between 20, 16 and 44, 16 should be maximum
    # as there is more weight when two objects boundaries overlap
    y_coord = params[0][1]
    for x_coord in range(params[0][0] + radius, params[1][0] - radius):
        distance_near_intersection = weight_map[x_coord, y_coord]
        nose.tools.assert_equal(max_weight_map, distance_near_intersection)



import logging
import os
from pathlib import Path
from typing import Literal, Optional, Sequence

import numpy as np
import torch
from iohub.ngff import ImageArray, Plate, Position, TransformationMeta, open_ome_zarr
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import BasePredictionWriter
from numpy.typing import DTypeLike, NDArray

from viscy.data.hcs import HCSDataModule, Sample

__all__ = ["HCSPredictionWriter"]
_logger = logging.getLogger("lightning.pytorch")


def _pad_shape(shape: tuple[int, ...], target: int = 5) -> tuple[int, ...]:
    """
    Pad shape tuple to a target length.
    Vendored from ``iohub.ngff.nodes._pad_shape()``.
    """
    pad = target - len(shape)
    return (1,) * pad + shape


def _resize_image(image: ImageArray, t_index: int, z_slice: slice) -> None:
    """Resize image array if incoming (1, C, Z, Y, X) stack is not within bounds."""
    if image.shape[0] <= t_index or image.shape[2] < z_slice.stop:
        _logger.debug(
            f"Resizing image '{image.name}' {image.shape} for "
            f"T={t_index}, Z-sclice={z_slice}."
        )
        image.resize(
            max(t_index + 1, image.shape[0]),
            image.channels,
            max(z_slice.stop, image.shape[2]),
            *image.shape[-2:],
        )


def _blend_in(old_stack: NDArray, new_stack: NDArray, z_slice: slice) -> NDArray:
    """
    Blend a new stack of images into an old stack over a specified range of slices.

    This function blends the `new_stack` of images into the `old_stack` over the range
    specified by `z_slice`. The blending is done using a weighted average where the
    weights are determined by the position within the range of slices. If the start
    of `z_slice` is 0, the function returns the `new_stack` unchanged.

    Parameters:
    ----------
    old_stack : NDArray
        The original stack of images to be blended.
    new_stack : NDArray
        The new stack of images to blend into the original stack.
    z_slice : slice
        A slice object indicating the range of slices over which to perform the blending.
        The start and stop attributes of the slice determine the range.

    Returns:
    -------
    NDArray
        The blended stack of images. If `z_slice.start` is 0, returns `new_stack` unchanged.
    """

    if z_slice.start == 0:
        return new_stack
    depth = z_slice.stop - z_slice.start
    # relevant predictions to integrate
    samples = min(z_slice.start + 1, depth)
    factors = []
    for i in reversed(list(range(depth))):
        factors.append(min(i + 1, samples))
    _logger.debug(f"Blending with factors {factors}.")
    factors = np.array(factors)[np.newaxis :, np.newaxis, np.newaxis]
    return old_stack * (factors - 1) / factors + new_stack / factors


class HCSPredictionWriter(BasePredictionWriter):
    """Callback to store virtual staining predictions as HCS OME-Zarr.

    :param str output_store: Path to the zarr store to store output
    :param bool write_input: Write the source and target channels too
        (must be writing to a new store),
        defaults to False
    :param Literal['batch', 'epoch', 'batch_and_epoch'] write_interval:
        When to write, defaults to "batch"
    """

    def __init__(
        self,
        output_store: str,
        write_input: bool = False,
        write_interval: Literal["batch", "epoch", "batch_and_epoch"] = "batch",
    ) -> None:
        super().__init__(write_interval)
        self.output_store = output_store
        self.write_input = write_input
        self._dataset_scale = None

    def _get_scale_metadata(self, metadata_store: Path) -> None:
        # Update the scale metadata
        if metadata_store is not None:
            with open_ome_zarr(metadata_store, mode="r") as metadata_store:
                if isinstance(metadata_store, Position):
                    self._dataset_scale = [
                        TransformationMeta(type="scale", scale=metadata_store.scale)
                    ]
                elif isinstance(metadata_store, Plate):
                    for _, pos in metadata_store.positions():
                        self._dataset_scale = [
                            TransformationMeta(type="scale", scale=pos.scale)
                        ]
                        break
                _logger.debug(f"Dataset scale {self._dataset_scale}.")

    def on_predict_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        dm: HCSDataModule = trainer.datamodule
        self._get_scale_metadata(dm.data_path)
        self.z_padding = dm.z_window_size // 2 if dm.target_2d else 0
        _logger.debug(f"Setting Z padding to {self.z_padding}")
        source_channel = dm.source_channel
        target_channel = dm.target_channel
        prediction_channel = [ch + "_prediction" for ch in target_channel]
        if os.path.exists(self.output_store):
            if self.write_input:
                raise FileExistsError(
                    "Cannot write input to an existing store. Aborting."
                )
            else:
                with open_ome_zarr(self.output_store, mode="r+") as plate:
                    for _, pos in plate.positions():
                        for ch in prediction_channel:
                            pos.append_channel(ch, resize_arrays=True)
                self.plate = open_ome_zarr(self.output_store, mode="r+")
        else:
            channel_names = prediction_channel
            if self.write_input:
                channel_names = source_channel + channel_names
            self.plate = open_ome_zarr(
                self.output_store, layout="hcs", mode="a", channel_names=channel_names
            )
        _logger.info(f"Writing prediction to: '{self.plate.zgroup.store.path}'.")
        if self.write_input:
            self.source_index = self._get_channel_indices(source_channel)
            self.target_index = self._get_channel_indices(target_channel)
        self.prediction_index = self._get_channel_indices(prediction_channel)

    def _get_channel_indices(self, channel_names: list[str]) -> list[int]:
        return [self.plate.get_channel_index(ch) for ch in channel_names]

    def write_on_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: torch.Tensor,
        batch_indices: Optional[Sequence[int]],
        batch: Sample,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        _logger.debug(f"Writing batch {batch_idx}.")
        for sample_index, _ in enumerate(batch["index"][0]):
            self.write_sample(batch, prediction[sample_index], sample_index)

    def on_predict_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.plate.close()

    def write_sample(
        self, batch: Sample, sample_prediction: torch.Tensor, sample_index: int
    ) -> None:
        _logger.debug(f"Writing sample {sample_index}.")
        sample_prediction = sample_prediction.cpu().numpy()
        img_name, t_index, z_index = [batch["index"][i][sample_index] for i in range(3)]
        t_index = int(t_index)
        z_index = int(z_index)
        # account for lost slices in 2.5D
        z_index += self.z_padding
        z_slice = slice(z_index, z_index + sample_prediction.shape[-3])
        image = self._create_image(
            img_name, sample_prediction.shape, sample_prediction.dtype
        )
        _resize_image(image, t_index, z_slice)
        if self.write_input:
            source_stack = batch["source"][sample_index].cpu()
            center_slice_index = source_stack.shape[-3] // 2
            image[t_index, self.source_index, z_index] = source_stack[
                :, center_slice_index
            ]
            if "target" in batch:
                image[t_index, self.target_index, z_index] = batch["target"][
                    sample_index
                ][:, center_slice_index].cpu()
        # write CZYX
        if self.z_padding == 0 and sample_prediction.shape[-3] > 1:
            old_stack = image.oindex[t_index, self.prediction_index, z_slice]
            sample_prediction = _blend_in(old_stack, sample_prediction, z_slice)
        image.oindex[t_index, self.prediction_index, z_slice] = sample_prediction

    def _create_image(self, img_name: str, shape: tuple[int], dtype: DTypeLike):
        if img_name in self.plate.zgroup:
            return self.plate[img_name]
        _logger.debug(f"Creating image '{img_name}'")
        _, row_name, col_name, pos_name, arr_name = img_name.split("/")
        position = self.plate.create_position(row_name, col_name, pos_name)
        shape = [1] + list(shape)
        shape[1] = len(position.channel_names)
        return position.create_zeros(
            arr_name,
            shape=shape,
            dtype=dtype,
            chunks=_pad_shape(tuple(shape[-2:]), 5),
            transform=self._dataset_scale,
        )



# %%
"""
Script to visualize the effective receptive field.
Original paper:
https://doi.org/10.48550/arXiv.1701.04128 (Luo et al., 2017).
"""

# %%
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
import torch

# from dlmbl_unet.unet import UNet
from viscy.light.engine import VSUNet

device = "cuda"


# %%
def plot_response(model: torch.nn.Module, title: str, ndim: Literal[2, 3] = 3) -> None:
    impulse_pattern = torch.zeros(1, 1, 800, 800, device=device)
    impulse_pattern[0, 0, 400, 400] = torch.finfo(torch.float32).eps
    if ndim == 3:
        impulse_pattern = impulse_pattern.unsqueeze(0)
    impulse = impulse_pattern.clone().requires_grad_(True)

    model.zero_grad()
    input_image = torch.zeros_like(impulse, requires_grad=True)
    fake_response = model(input_image)
    fake_loss = (fake_response * impulse_pattern.clone()).sum()
    fake_loss.backward()
    response_backwards = input_image.grad.detach().cpu().numpy()
    if ndim == 3:
        response_backwards = response_backwards.squeeze(0)
    scaled_response = response_backwards[0, 0] / response_backwards.max()

    f, ax = plt.subplots(1, 2, figsize=(10, 5))
    spectrum = np.log1p(np.abs(scaled_response))
    im = ax[0].imshow(spectrum, cmap="nipy_spectral")
    _ = plt.colorbar(im, label="log power", shrink=0.5, location="bottom")
    ax[1].plot(np.log1p(scaled_response.std(axis=(0))), label="rows", alpha=0.5)
    ax[1].plot(np.log1p(scaled_response.std(axis=(1))), label="columns", alpha=0.5)
    ax[1].legend()
    ax[1].set_ylabel("log std of power")
    f.suptitle(title)
    f.tight_layout()


# %%
# unet = UNet(depth=4, in_channels=1).to(device)

# plot_response(unet, "DLMBL U-Net (initialization)", ndim=2)

# %%
model_kwargs = dict(
    architecture="fcmae",
    model_config=dict(
        in_channels=1,
        out_channels=2,
        encoder_blocks=[3, 3, 9, 3],
        dims=[96, 192, 384, 768],
        decoder_conv_blocks=2,
        stem_kernel_size=(1, 2, 2),
        in_stack_depth=1,
        pretraining=False,
    ),
)

initial_model = VSUNet(**model_kwargs).to(device)
trained_model = VSUNet.load_from_checkpoint(
    "/hpc/websites/public.czbiohub.org/comp.micro/viscy/VS_models/VSCyto2D/VSCyto2D/epoch=399-step=23200.ckpt",
    **model_kwargs,
).to(device)

plot_response(initial_model, "UNeXt2 (initialization)")
plot_response(trained_model, "UNeXt2 (trained)")

# %%



# %%
from pathlib import Path

import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.preprocessing import StandardScaler

from viscy.representation.embedding_writer import read_embedding_dataset
from viscy.representation.evaluation.clustering import (
    compare_time_offset,
    pairwise_distance_matrix,
    rank_nearest_neighbors,
    select_block,
)

# %%
prediction_path = Path(
    "/hpc/projects/organelle_phenotyping/ALFI_benchmarking/predictions_final/ALFI_opp_7mins.zarr"
)

embeddings = read_embedding_dataset(prediction_path)
features = embeddings["features"]

# %%
scaled_features = StandardScaler().fit_transform(features.values)

# %%
cross_dist = pairwise_distance_matrix(scaled_features, metric="cosine")
rank_fractions = rank_nearest_neighbors(cross_dist, normalize=True)

# %%
# select a single track in a single fov
fov = "/0/0/0"
fov_idx = (features["fov_name"] == fov).values

track_id = 1
track_idx = (features["track_id"] == track_id).values

fov_and_track_idx = fov_idx & track_idx

single_track_dissimilarity = select_block(cross_dist, fov_and_track_idx)
single_track_rank_fraction = select_block(rank_fractions, fov_and_track_idx)

piece_wise_dissimilarity = compare_time_offset(
    single_track_dissimilarity, time_offset=1
)
piece_wise_rank_difference = compare_time_offset(
    single_track_rank_fraction, time_offset=1
)

# %%
f = plt.figure(figsize=(8, 12))
f.suptitle(f"Track {track_id} in FOV {fov}")
subfigs = f.subfigures(2, 1, height_ratios=[1, 2])

umap = subfigs[0].subplots(1, 1)
single_cell_features = features.sel(fov_name=fov, track_id=track_id).sortby("t")
sns.lineplot(
    x=single_cell_features["UMAP1"],
    y=single_cell_features["UMAP2"],
    ax=umap,
    color="k",
    alpha=0.5,
)
sns.scatterplot(
    x=features["UMAP1"], y=features["UMAP2"], ax=umap, color="k", s=100, alpha=0.01
)
sns.scatterplot(
    x=single_cell_features["UMAP1"],
    y=single_cell_features["UMAP2"],
    hue=single_cell_features["t"],
    ax=umap,
    palette="RdYlGn",
)

f1 = subfigs[1]
ax = f1.subplots(2, 2)

sns.heatmap(single_track_dissimilarity, ax=ax[0, 0], square=True)
ax[0, 0].set_title("Cosine dissimilarity")
ax[0, 0].set_xlabel("Frame")
ax[0, 0].set_ylabel("Frame")

sns.heatmap(single_track_rank_fraction, ax=ax[0, 1], square=True)
ax[0, 1].set_title("Column-wise normalized neighborhood distance")
ax[0, 1].set_xlabel("Frame")
ax[0, 1].set_ylabel("Frame")

sns.lineplot(piece_wise_dissimilarity, ax=ax[1, 0])
ax[1, 0].set_title("$1 - \cos{(t_i, t_{i+1})}$")
ax[1, 0].set_xlabel("Frame")
ax[1, 0].set_ylabel("Cosine dissimilarity")

sns.lineplot(piece_wise_rank_difference, ax=ax[1, 1])
ax[1, 1].set_title("Nearest neighbor fraction difference")
ax[1, 1].set_xlabel("Frame")
ax[1, 1].set_ylabel("Rank fraction")

# %%



from __future__ import annotations

from abc import ABC, abstractmethod
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
import torch
from iohub.ngff import Plate, Position, open_ome_zarr
from lightning.pytorch import LightningDataModule
from monai.data.meta_obj import set_track_meta
from monai.data.utils import list_data_collate
from monai.transforms.compose import Compose
from torch import Tensor
from torch.multiprocessing import Manager
from torch.utils.data import DataLoader, Dataset

from viscy.data.distributed import ShardedDistributedSampler
from viscy.data.hcs import _ensure_channel_list, _read_norm_meta
from viscy.data.select import SelectWell
from viscy.data.typing import DictTransform, NormMeta

if TYPE_CHECKING:
    from multiprocessing.managers import DictProxy

_logger = getLogger("lightning.pytorch")

_CacheMetadata = tuple[Position, int, NormMeta | None]


class GPUTransformDataModule(ABC, LightningDataModule):
    """Abstract data module with GPU transforms."""

    train_dataset: Dataset
    val_dataset: Dataset
    batch_size: int
    num_workers: int
    pin_memory: bool
    prefetch_factor: int | None

    def _maybe_sampler(
        self, dataset: Dataset, shuffle: bool
    ) -> ShardedDistributedSampler | None:
        return (
            ShardedDistributedSampler(dataset, shuffle=shuffle)
            if torch.distributed.is_initialized()
            else None
        )

    def train_dataloader(self) -> DataLoader:
        sampler = self._maybe_sampler(self.train_dataset, shuffle=True)
        _logger.debug(f"Using training sampler {sampler}")
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=False if sampler else True,
            sampler=sampler,
            persistent_workers=True if self.num_workers > 0 else False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=False,
            collate_fn=list_data_collate,
            prefetch_factor=self.prefetch_factor,
        )

    def val_dataloader(self) -> DataLoader:
        sampler = self._maybe_sampler(self.val_dataset, shuffle=False)
        _logger.debug(f"Using validation sampler {sampler}")
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            sampler=sampler,
            persistent_workers=True if self.num_workers > 0 else False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=False,
            collate_fn=list_data_collate,
            prefetch_factor=self.prefetch_factor,
        )

    @property
    @abstractmethod
    def train_cpu_transforms(self) -> Compose: ...

    @property
    @abstractmethod
    def train_gpu_transforms(self) -> Compose: ...

    @property
    @abstractmethod
    def val_cpu_transforms(self) -> Compose: ...

    @property
    @abstractmethod
    def val_gpu_transforms(self) -> Compose: ...


class CachedOmeZarrDataset(Dataset):
    """Dataset for cached OME-Zarr arrays.

    Parameters
    ----------
    positions : list[Position]
        List of FOVs to load images from.
    channel_names : list[str]
        List of channel names to load.
    cache_map : DictProxy
        Shared dictionary for caching loaded volumes.
    transform : Compose | None, optional
        Composed transforms to be applied on the CPU, by default None
    array_key : str, optional
        The image array key name (multi-scale level), by default "0"
    load_normalization_metadata : bool, optional
        Load normalization metadata in the sample dictionary, by default True
    skip_cache : bool, optional
        Skip caching to save RAM, by default False
    """

    def __init__(
        self,
        positions: list[Position],
        channel_names: list[str],
        cache_map: DictProxy,
        transform: Compose | None = None,
        array_key: str = "0",
        load_normalization_metadata: bool = True,
        skip_cache: bool = False,
    ):
        key = 0
        self._metadata_map: dict[int, _CacheMetadata] = {}
        for position in positions:
            img = position[array_key]
            norm_meta = _read_norm_meta(position)
            for time_idx in range(img.frames):
                cache_map[key] = None
                self._metadata_map[key] = (position, time_idx, norm_meta)
                key += 1
        self.channels = {ch: position.get_channel_index(ch) for ch in channel_names}
        self.array_key = array_key
        self._cache_map = cache_map
        self.transform = transform
        self.load_normalization_metadata = load_normalization_metadata
        self.skip_cache = skip_cache

    def __len__(self) -> int:
        return len(self._metadata_map)

    def __getitem__(self, idx: int) -> dict[str, Tensor]:
        position, time_idx, norm_meta = self._metadata_map[idx]
        cache = self._cache_map[idx]
        if cache is None:
            _logger.debug(f"Loading volume for index {idx}")
            volume = torch.from_numpy(
                position[self.array_key]
                .oindex[time_idx, list(self.channels.values())]
                .astype(np.float32)
            )
            if not self.skip_cache:
                _logger.debug(f"Caching for index {idx}")
                self._cache_map[idx] = volume
        else:
            _logger.debug(f"Using cached volume for index {idx}")
            volume = cache
        sample = {name: img[None] for name, img in zip(self.channels.keys(), volume)}
        if self.load_normalization_metadata:
            sample["norm_meta"] = norm_meta
        if self.transform:
            sample = self.transform(sample)
        if not isinstance(sample, list):
            sample = [sample]
        return sample


class CachedOmeZarrDataModule(GPUTransformDataModule, SelectWell):
    """Data module for cached OME-Zarr arrays.

    Parameters
    ----------
    data_path : Path
        Path to the HCS OME-Zarr dataset.
    channels : str | list[str]
        Channel names to load.
    batch_size : int
        Batch size for training and validation.
    num_workers : int
        Number of workers for data-loaders.
    split_ratio : float
        Fraction of the FOVs used for the training split.
        The rest will be used for validation.
    train_cpu_transforms : list[DictTransform]
        Transforms to be applied on the CPU during training.
    val_cpu_transforms : list[DictTransform]
        Transforms to be applied on the CPU during validation.
    train_gpu_transforms : list[DictTransform]
        Transforms to be applied on the GPU during training.
    val_gpu_transforms : list[DictTransform]
        Transforms to be applied on the GPU during validation.
    pin_memory : bool, optional
        Use page-locked memory in data-loaders, by default True
    skip_cache : bool, optional
        Skip caching for this dataset, by default False
    include_wells : list[str], optional
        List of well names to include in the dataset, by default None (all)
    include_wells : list[str], optional
        List of well names to include in the dataset, by default None (all)
    """

    def __init__(
        self,
        data_path: Path,
        channels: str | list[str],
        batch_size: int,
        num_workers: int,
        split_ratio: float,
        train_cpu_transforms: list[DictTransform],
        val_cpu_transforms: list[DictTransform],
        train_gpu_transforms: list[DictTransform],
        val_gpu_transforms: list[DictTransform],
        pin_memory: bool = True,
        skip_cache: bool = False,
        include_wells: list[str] | None = None,
        exclude_fovs: list[str] | None = None,
    ):
        super().__init__()
        self.data_path = data_path
        self.channels = _ensure_channel_list(channels)
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.split_ratio = split_ratio
        self._train_cpu_transforms = Compose(train_cpu_transforms)
        self._val_cpu_transforms = Compose(val_cpu_transforms)
        self._train_gpu_transforms = Compose(train_gpu_transforms)
        self._val_gpu_transforms = Compose(val_gpu_transforms)
        self.pin_memory = pin_memory
        self.skip_cache = skip_cache
        self._include_wells = include_wells
        self._exclude_fovs = exclude_fovs

    @property
    def train_cpu_transforms(self) -> Compose:
        return self._train_cpu_transforms

    @property
    def train_gpu_transforms(self) -> Compose:
        return self._train_gpu_transforms

    @property
    def val_cpu_transforms(self) -> Compose:
        return self._val_cpu_transforms

    @property
    def val_gpu_transforms(self) -> Compose:
        return self._val_gpu_transforms

    def _set_fit_global_state(self, num_positions: int) -> list[int]:
        # disable metadata tracking in MONAI for performance
        set_track_meta(False)
        # shuffle positions, randomness is handled globally
        return torch.randperm(num_positions).tolist()

    def _include_well_name(self, name: str) -> bool:
        if self._include_wells is None:
            return True
        else:
            return name in self._include_wells

    def _filter_fit_fovs(self, plate: Plate) -> list[Position]:
        positions = []
        for well_name, well in plate.wells():
            if self._include_well_name(well_name):
                for _, p in well.positions():
                    positions.append(p)
        if len(positions) < 2:
            raise ValueError(
                "At least 2 FOVs are required for training and validation."
            )
        return positions

    def setup(self, stage: Literal["fit", "validate"]) -> None:
        if stage not in ("fit", "validate"):
            raise NotImplementedError("Only fit and validate stages are supported.")
        cache_map = Manager().dict()
        plate: Plate = open_ome_zarr(self.data_path, mode="r", layout="hcs")
        positions = self._filter_fit_fovs(plate)
        shuffled_indices = self._set_fit_global_state(len(positions))
        num_train_fovs = int(len(positions) * self.split_ratio)
        train_fovs = [positions[i] for i in shuffled_indices[:num_train_fovs]]
        val_fovs = [positions[i] for i in shuffled_indices[num_train_fovs:]]
        _logger.debug(f"Training FOVs: {[p.zgroup.name for p in train_fovs]}")
        _logger.debug(f"Validation FOVs: {[p.zgroup.name for p in val_fovs]}")
        self.train_dataset = CachedOmeZarrDataset(
            train_fovs,
            self.channels,
            cache_map,
            transform=self.train_cpu_transforms,
            skip_cache=self.skip_cache,
        )
        self.val_dataset = CachedOmeZarrDataset(
            val_fovs,
            self.channels,
            cache_map,
            transform=self.val_cpu_transforms,
            skip_cache=self.skip_cache,
        )
